######################################################################################
# External imports

import numpy

######################################################################################
# Internal imports

from galapy.sampling.Results import Results
from galapy.Handlers import ModelParameters 
from galapy.Galaxy import gxy_params_defaults 
from galapy.Noise import noise_params_defaults
param_defaults = dict( **{'.'.join(['galaxy',k]):v for k,v in gxy_params_defaults.items()},
                       **{'.'.join(['noise',k]):v for k,v in noise_params_defaults.items()} )

######################################################################################

def get_parameters_summary_statistics ( res, stat_type = 'quantiles', quantile = None ) :
    """ Function returning a dictionary with summary statistics of the samples in 
    the Results instance of some sampling run
    
    Parameters
    ----------
    res : galapy.sampling.Results.Results
        An instance of type ``Results``
    stat_type : str
        The desired summary statistics. 
        Available statistics are
        'bestfit_and_interval', 'mean_and_std' and 'quantiles'
    quantile : scalar or sequence
        If ``stat_type='bestfit_and_interval'``, the probability
        enclosed in the credible interval. Default is 0.68, 
        i.e. 68% credible interval.
        If ``stat_type='quantiles'``, the quantiles requested.
        Default is (0.16,0.5,0.84), i.e. median and 68% bound.
    
    Returns
    -------
    : ndarray
        - if ``stat_type='bestfit_and_interval'`` each row is an array with size = 3 
          where the first element is the best-fit value, the second and third are the 
          lower and upper uncertainties, respectively.
        - if ``stat_type='quantiles'`` each row is an array with size = ``len(quantile)``
        - if ``stat_type='mean_and_std'`` each row is an array of size = 2 where the
          first element is the mean and the second element is the standard deviation.
    """
    
    if stat_type not in ('bestfit_and_interval', 'mean_and_std', 'quantiles') :
        raise ValueError(
            "Allowed values for argument ``stat_type`` are "
            "'bestfit_and_interval', 'mean_and_std' and 'quantiles'"
        )

    if stat_type == 'bestfit_and_interval' :
        if quantile is None :
            print( 'Falling back to default percentile: 0.68')
            quantile = 0.68
        try :
            quantile = float( quantile )
        except :
            raise
        summary = numpy.empty([self.ndim,3], dtype=float)
        idcentre = self.logl[self.wnot0].argmax()
        for i, sample in enumerate( self.samples.T ) :
            summary[i,0] = sample[idcentre]
            low, upp = get_credible_interval(
                sample, idcentre, quantile, self.weights[self.wnot0]
            )
            if low is None : low = -numpy.inf
            if upp is None : upp = +numpy.inf
            summary[i,1] = low
            summary[i,2] = upp
        
    if stat_type == 'quantiles' :
        if quantile is None or len(quantile)!=3:
            print( 'Falling back to default quantiles: (0.16,0.5,0.84)')
            quantile = [0.16, 0.5, 0.84]
        summary = res.get_quantile('samples', quantile).T
        summary[:,0] = abs(summary[:,0]-summary[:,1])
        summary[:,2] = abs(summary[:,2]-summary[:,1])
        
    if stat_type == 'mean_and_std' :
        summary = numpy.array([res.get_mean('samples'), 
                               res.get_std('samples')]).T
    return summary

######################################################################################

def get_parameters_summary_strings ( res, digits = 2,
                                     stat_type = 'quantiles', quantile = None ) :
    """ Function returning a list of formatted strings with the chosen summary 
    statistics of the samples in the Results instance of some sampling run
    
    Parameters
    ----------
    res : galapy.sampling.Results.Results
        An instance of type ``Results``
    digits : integer
        Number of significant digits to show.
        Each statistic will be represented in strings formatted as ``{:.{digits}f}``
        Default is 2
    stat_type : str
        The desired summary statistics. Available statistics are
        'mean_and_std' and 'quantiles'
    quantile : scalar or sequence
        The quantiles requested (used only is ``stat_type='quantiles'``.
        Default is (0.16,0.5,0.84), i.e. median and 68% bound.
    
    Returns
    -------
    : ndarray
        An array with Ndim TeX-formatted summary statistics, where Ndim is the
        size of the free-parameters space of the sampling run.
    """
    
    summary = get_parameters_summary_statistics( res, stat_type, quantile )
    fstring = None
    if stat_type == 'quantiles' :
        fstring = f'{{1:.{digits}f}}_{{{{-{{0:.{digits}f}}}}}}^{{{{+{{2:.{digits}f}}}}}}'
    if stat_type == 'mean_and_std' :
        fstring = f'{{0:.{digits}f}} \\pm {{1:.{digits}f}}'
    return numpy.array([fstring.format(*s) for s in summary])

######################################################################################

def get_parameters_label_strings ( handler ) :
    """ Returns strings formatted for TeX math
    
    Parameters
    ----------
    handler : galapy.Handlers.ModelParameters
        an instance of type ``ModelParameters``
    
    Returns
    -------
    : dict
        Dictionary where the keys are the name of the free-parameter
        and the values are the TeX-formatted strings
    """

    if not isinstance( handler, ModelParameters ) :
        raise ValueError(
            "Attribute ``handler`` should be an instance of type ``ModelParameters``"
        )
    
    labels = { 
        key : param_defaults[key][3] 
        if not log else '\\log~'+param_defaults[key][3]
        for key, log in zip(handler.par_free, 
                            handler.par_log) 
    }

    return labels

######################################################################################



""" The Inter-Stellar Medium module implements and wraps the absorbing and re-radiating media of the Galaxy.
"""

# External Imports
import numpy

# Internal imports
from .ISM_core import CMC, CDD, total_attenuation

ism_tunables = {
    # Molecular Clouds
    'mc' : [ 'f_MC', 'norm_MC', 'N_MC', 'R_MC', 'Zgas', 'tau_esc', 'Mgas', 'dMClow', 'dMCupp' ],
    
    # Diffuse Dust
    'dd' : [ 'f_MC', 'norm_DD', 'Mdust', 'Rdust', 'f_PAH', 'dDDlow', 'dDDupp' ],
}
""" Dictionary with the tunable parameters of the ISM phases

#. :code:`mc` : **Molecular Clouds**
#. :code:`dd` : **Diffuse Dust**

"""

def ism_build_params ( phase, **kwargs ) :
    """ Builds the parameters dictionary for given phase.
    """
    if phase not in ism_tunables.keys() :
        av_phases = ''
        for k in ism_tunables.keys() :
            av_phases += f'"{k}" '
        raise ValueError( f'Phase "{phase}" chosen not available. '
                          f'Available phases are: {av_phases}')
    if phase == 'mc' :
        out = {
            'f_MC'    : 0.5,
            'norm_MC' : 1.e+02,
            'N_MC'    : 1.e+03,
            'R_MC'    : 10.,
            'Zgas'    : 0.01,
            'tau_esc' : 1.e+07,
            'Mgas'    : 1.e+09,
            'dMClow'  : 1.3,
            'dMCupp'  : 1.6,
        }
        
    if phase == 'dd' :
        out = {
            'f_MC'    : 0.5,
            'norm_DD' : 1.e+00,
            'Mdust'   : 1.e+07,
            'Rdust'   : 1.e+03,
            'f_PAH'   : 0.2,
            'dDDlow'  : 0.7,
            'dDDupp'  : 2.0,
        }
        
    for k in set( out.keys() ).intersection(kwargs.keys()) :
        out[k] = kwargs[k]
    return out
    
class ismPhase ():
    """ ISM phase base class.
    """

    def __init__ ( self, phase, builder, T = None, **kwargs ) :
        
        self.phase  = phase
        self.params = ism_build_params( self.phase, **kwargs )
        self.core   = builder()
        self.set_parameters( **kwargs )
        self.T = None
        if T :
            self.set_temperature( T )
                
        # steal docstrings from C-core:
        self.set_temperature.__func__.__doc__ = self.core.set_temperature.__doc__
        self.temperature.__func__.__doc__     = self.core.temperature.__doc__
        self.emission.__func__.__doc__        = self.core.emission.__doc__
        self.attenuation.__func__.__doc__     = self.core.attenuation.__doc__
        self.extinction.__func__.__doc__      = self.core.extinction.__doc__
        self.A_V.__func__.__doc__             = self.core.A_V.__doc__

    def set_parameters ( self, **kwargs ) :
        """ Function for setting the free parameters of the class.

        It is implemented to take any number of keyword arguments and
        automathically ignores arguments that are not free-parameters of the class.
        """
        self.params.update( kwargs )
        self.core.set_params( numpy.asarray( [
            self.params[k]
            for k in ism_tunables[self.phase]
        ], dtype = float ) )
    
    def set_temperature ( self, T ) :
        """ Manually set the temperature of the ISM phase
        
        Parameters
        ----------
        T : float
          Temperature in Kelvin :math:`[K]`
        """
        self.T = T
        self.core.set_temperature( self.T )
        return;
    
    def set_slopes ( self, lower, upper ) :
        r""" Manually set the slopes of the ISM phase extinction
        
        Parameters
        ----------
        lower : float
          slope of extinction at wavelengths <= 100 :math:`\mu m`
        lower : float
          slope of extinction at wavelengths > 100 :math:`\mu m`
          
        """
        self.params[ism_tunables[self.phase][-2]] = lower 
        self.params[ism_tunables[self.phase][-1]] = upper
        self.core.set_slopes( lower, upper )
        return;

    def temperature ( self, Etot ) :
        r""" Computes the temperature of the ISM at given total energy.

        With the assumption that the total energy density :math:`E_\text{abs}^\text{phase}`
        absorbed by the ISM phase is re-radiated as a grey body 
        with luminosity :math:`L_\lambda[\tau, T]`,
        the ISM temperature is computed by solving the integral
        
        .. math:: 
          
          \int_0^\infty \text{d}\lambda L_\lambda^\text{phase}[\tau|T_\text{phase}(\tau)] =
          E_\text{abs}^\text{phase}(\tau)
        
        Parameters
        ----------
        Etot : scalar float
          Total energy absorbed by the ISM phase.

        Returns
        -------
        T : scalar float
          Temperature in Kelvin :math:`[K]` of the given ISM phase.
        """
        self.T = self.core.temperature( Etot )
        return self.T

    def emission ( self, wavelength, T = None ) :
        r""" Computes the ISM emission at given wavelength.
        
        We assume the ISM radiates as a gray body with emission spectrum
        
        .. math::
        
          L_\lambda^\text{phase}(\tau) = 
          \mathcal{N}_\text{phase}\; \bigl[1 
          - 10^{-0.4\,A_\lambda^\text{phase}(\tau)}\bigr]\;
          B_\lambda(T_\text{phase})
        
        where :math:`\mathcal{N}_\text{phase}` is a normalization depending
        on the model parameters, 
        the factor :math:`1 - 10^{-0.4\,A_\lambda^\text{phase}(\tau)}` 
        is the optical depth of the ISM phase and :math:`B_\lambda(T_\text{phase})`
        is the temperature of the medium.          

        Parameters
        ----------
        wavelength : array or scala float
          wavelength in angstroms :math:`[\mathring{A}]`

        Keyword Arguments
        -----------------
        T : float
          temperature in Kelvin :math:`[K]` of the ISM phase 
          (optional, default = :code:`None`)

        Returns
        -------
        L_ISM : float of array-like of floats
          Luminosity at given wavelength 
        """
        if T :
            self.set_temperature( T )
        return self.core.emission( wavelength )

    def attenuation ( self, wavelength ) :
        """ Computes the ISM attenuation at given wavelength.

        Parameters
        ----------
        wavelength : array or scalar float
          wavelength in angstroms :math:`[\mathring{A}]`
        
        Returns
        -------
         : array or scalar float
        """
        return self.core.attenuation( wavelength )

    def extinction ( self, wavelength ) :
        """ Computes the ISM extinction at given wavelength.

        Parameters
        ----------
        wavelength : array or scalar float
          wavelength in angstroms :math:`[\mathring{A}]`
        
        Returns
        -------
         : array or scalar float
        """
        return self.core.extinction( wavelength )

    def A_V ( self ) :
        """ Returns the extinction value in the visible band.
        """
        return self.core.A_V()

class MC ( ismPhase ) :
    """Class implementing the Molecular-Cloud medium.
    
    Parameters
    ----------
    T : float
       (optional, default = None) average temperature of Molecular Clouds 
    kwargs : dictionary
       (optional) can contain any parameter value (key-value pairs)
    """

    def __init__ ( self, T = None, **kwargs ) :
        super().__init__( 'mc', CMC, T=T, **kwargs )

    def eta ( self, tt ) :
        r"""Implements the function regulating the time-dependent
        evaporation of Molecular clouds:

        .. math::
            
            \eta(\tau) 
            = \begin{cases}
                1  & \tau\leq \tau_\text{esc}\\
                2-\frac{\tau}{\tau_\text{esc}} & \tau_\text{esc}<\tau\leq 2\,\tau_\text{esc}\\
                0 & \tau>2\,\tau_\text{esc}
            \end{cases}

        where :math:`\tau_\text{esc}` is a free parameter of the MC class and :math:`\tau` is the
        age at which the hosting galaxy is being observed.

        Parameters
        ----------
        tt : array or scalar float
            age of the hosting galaxy (in years)
        
        Returns
        -------
        : array or scalar float
           value of the :math:`\eta(\tau)` parameter at the input age(s)
        """
        return self.core.eta( tt )

    def time_attenuation ( self, wavelength, tt ) :
        r"""Computes the time-dependent attenuation due to Molecular Clouds:
        
        .. math::
            
            \mathcal{A}_\text{MC}(\lambda) = 1-\eta(\tau)\left[ 1 - 10^{-0.4\, A_\text{MC}(\lambda)} \right]
        
        where :math:`\eta(\tau)` is computed with :code:`MC.eta()`.
        
        Parameters
        ----------
        wavelength : array or scalar float
            wavelength in angstroms :math:`[\mathring{A}]`
        tt : array or scalar float
            age of the hosting galaxy (in years)
        
        Returns
        -------
        : float or iterable
            Output value of the time-dependent attenutation from MCs. The output shape depends
            on the inputs and is = :code:`(len(wavelength), len(tt))`
        """
        return (
            1 - ( 1 - self.attenuation( wavelength ) )[:,numpy.newaxis]
            * self.eta( tt )[numpy.newaxis,:]
        )
        
class DD ( ismPhase ) :
    """Class implementing the Diffuse-Dust medium.
    
    Parameters
    ----------
    T : float
       (optional, default = None) average temperature of Diffuse Dust 
    kwargs : dictionary
       (optional) can contain any parameter value (key-value pairs)
    """

    def __init__ ( self, T = None, **kwargs ) :
        super().__init__( 'dd', CDD, T=T, **kwargs )

class ISM () :
    """Class implementing the Inter-Stellar Medium.
    Wraps up the combined effect of Molecular Clouds and Diffuse Dust.
    
    Parameters
    ----------
    TMC : float
       (optional, default = None) average temperature of Molecular Clouds 
    TDD : float
       (optional, default = None) average temperature of Diffuse Dust 
    kwargs : dictionary
       (optional) can contain any parameter value (key-value pairs) of both
       MCs and DD
    """

    def __init__ ( self, TMC = None, TDD = None, **kwargs ) :
        self.mc = MC( TMC,
                      **{ k : v
                          for k, v in kwargs.items()
                          if k in ism_tunables[ 'mc' ] } )
        self.dd = DD( TDD,
                      **{ k : v
                          for k, v in kwargs.items()
                          if k in ism_tunables[ 'dd' ] } )

    def set_parameters ( self, **kwargs ) :
        """Change the value of the free parameters.
        
        Keyword Arguments
        -----------------
        f_MC    : float
          Fraction of the total dust mass in MCs 
        norm_MC : float
          Normalization of the attenuation law of MCs
        N_MC    : float
          Number of MCs within the hosting galaxy
        R_MC    : float
          Average radius of a MC
        Zgas    : float
          Average metallicity of the gas in the hosting galaxy
        tau_esc : float
          Average escape time of SSPs from MCs
        Mgas    : float
          Total gas mass
        dMClow  : float
          spectral index of the MC extinction at short wavelengths
        dMCupp  : float
          spectral index of the MC extinction at long wavelengths
        norm_DD : float
          Normalization of the attenuation law of DD
        Mdust   : float
          Total dust mass
        Rdust   : float
          Radius of the DD region
        f_PAH   : float
          fraction of the total emission from DD contributed by PAH
        dDDlow  : float
          spectral index of the DD extinction at short wavelengths
        dDDupp  : float
          spectral index of the DD extinction at long wavelengths
        """
        self.mc.set_parameters( **{ k : v
                                    for k, v in kwargs.items()
                                    if k in ism_tunables[ 'mc' ] } )
        self.dd.set_parameters( **{ k : v
                                    for k, v in kwargs.items()
                                    if k in ism_tunables[ 'dd' ] } )
        return;

    def total_attenuation ( self, wavelength, tt ) :
        """Function computing the total attenuation due to MCs and DD.
        
        Note that the output is flattened and the total size depends on the
        size of the input arrays.
        Both output arrays (e.g. :code:`arr`) can be reshaped to their original dimensions by calling
        
        .. code::
            
            arr.reshape( wavelength.size, tt.size )
        
        for the case when both :code:`wavelength` and :code:`tt` are arrays.
        If one of them is not, flattening the output does not have any effect.
        
        Parameters
        ----------
        wavelength : array or scalar float
          wavelength in angstroms :math:`[\mathring{A}]`
        tt : array or scalar float
          age of the hosting galaxy (in years)
        
        Returns
        -------
        : 1d array
          Total time attenuation due to MCs for each wavelength and age in the input grid.
          (uses function :code:`MC.time_attenuation` and flattens the output array.
        : 1d array
          Total time attenuation due to the combined effect of MCs first and, subsequently, DD.
          (the output array is flattened)
        """
        attMC = self.mc.time_attenuation( wavelength, tt )
        attDD = attMC * self.dd.attenuation( wavelength )[:,numpy.newaxis]
        return ( numpy.ascontiguousarray( attMC.ravel() ),
                 numpy.ascontiguousarray( attDD.ravel() ) )




import numpy
import os
import warnings
os.environ["OMP_NUM_THREADS"] = "1"

import argparse
import importlib.util

from galapy.PhotometricSystem import PMS
from galapy.Galaxy import PhotoGXY
from galapy.Noise import CalibrationError
from galapy.Handlers import ModelParameters
from galapy.sampling.Statistics import gaussian_loglikelihood
from galapy.sampling.Sampler import Sampler
from galapy.sampling.Observation import Observation
from galapy.sampling.Results import generate_output_base, dump_results
from galapy.internal.utils import set_nested

_default_sampling_kw = {
    'dynesty' : dict(
        dlogz_init=0.05, nlive_init=500, nlive_batch=100,
        maxiter_init=10000, maxiter_batch=1000, maxbatch=10,
        stop_kwargs = {'target_n_effective': int(5e6)}
    ),
    'emcee' : {},
}    

global_dict = None

################################################################################

def initialize ( bands, fluxes, errors, uplims, filter_args, params,
                 sfh_model = 'insitu', ssp_lib = 'parsec22.NT',
                 do_Radio = False, do_Xray = False, do_AGN = False,
                 noise_model = None, noise_params = {},
                 gxy_kwargs = {}, noise_kwargs = {}, filter_kwargs = {} ) :

    global global_dict 
    
    #########################################################################
    # Build photometric system

    pms = PMS( *filter_args, **filter_kwargs )
    
    #########################################################################
    # Build observation

    data = Observation( bands, fluxes, errors, uplims, pms )
    
    #########################################################################
    # Build photometric galaxy

    model = PhotoGXY( pms = data.pms,
                      sfh = { 'model' : sfh_model },
                      csp = { 'ssp_lib' : ssp_lib },
                      do_Radio = do_Radio,
                      do_Xray = do_Xray,
                      do_AGN = do_AGN,
                      **gxy_kwargs )
    noise = None
    if noise_model is not None :
        if noise_model == 'calibration_error' :
            noise = CalibrationError( **noise_kwargs )
        else :
            warnings.warn( 'The noise model chosen is not valid, noise will be ignored' )
    
    #########################################################################
    # Build parameters handler

    sample_params = { '.'.join( ['galaxy', key] ) : value for key, value in params.items() }
    if noise is not None :
        sample_params.update( { '.'.join( ['noise', key] ) : value
                                for key, value in noise_params.items() } )
        handler = ModelParameters( model, noise,
                                   sample_params = sample_params )
    else :
        handler = ModelParameters( model, sample_params = sample_params )

    #########################################################################
    # Set fixed parameters and initial values for free parameters

    #init = {}
    #for klist, v in handler.parameters.items() :
    #    set_nested( init, klist.split('.'), v )
    init = handler.return_nested()
    try :
        init['galaxy']['age'] = min(
            model.cosmo.age( init['galaxy']['redshift'] ) - 1.0,
            init['galaxy']['age']
        )
        model.set_parameters( **init['galaxy'] )
    except RuntimeError :
        pass
    if noise is not None : noise.set_parameters( **init['noise'] )

    global_dict = { 'data' : data, 'model' : model, 'noise' : noise, 'handler' : handler }
    return;

################################################################################

def loglikelihood ( par, **kwargs ) :
    
    data    = global_dict['data']
    model   = global_dict['model']
    noise   = global_dict['noise']
    handler = global_dict['handler']

    nested = handler.return_nested( par )
    try :
        model.set_parameters( **nested['galaxy'] )
        flux_model = numpy.asarray( model.photoSED() )
    except RuntimeError :
        return -numpy.inf

    if noise is not None :
        noise.set_parameters( **nested['noise'] )
        errors = noise.apply( data.errors, flux_model )
        noise_llike = numpy.log( 2 * numpy.pi * errors**2 ).sum()
        if numpy.isnan( noise_llike ) :
            return - numpy.inf
        return gaussian_loglikelihood( data   = data.fluxes,
                                       error  = errors,
                                       model  = flux_model,
                                       uplims = data.uplims,
                                       **kwargs ) - 0.5 * noise_llike
    return gaussian_loglikelihood( data   = data.fluxes,
                                   error  = data.errors,
                                   model  = flux_model,
                                   uplims = data.uplims,
                                   **kwargs )

################################################################################

# Include prior in likelihood (here uniform prior)
def logprob ( par, **kwargs ) :
    
    pmin, pmax = global_dict['handler'].par_prior.T
    if all( ( pmin < par ) & ( par < pmax ) ) :
        return loglikelihood( par, **kwargs )
    
    return -numpy.inf

################################################################################

def sample ( sampler = 'emcee', nwalkers = None, nsamples = None, 
             sampler_kw = {}, logl_kw = {}, run_sampling_kw = {},
             Ncpu = 1, pool = None ) :
    
    # Sampler keywords
    sampler_kw = dict( sampler_kw )
    
    # Sampling keywords
    sampling_kw = dict(_default_sampling_kw[sampler])
    sampling_kw.update(run_sampling_kw)
    
    # It's run-time!
    if sampler == 'dynesty' :
        
        # Define prior-transform to map parameters in the unit-cube
        from galapy.sampling.Statistics import transform_to_prior_unit_cube
    
        # Build sampler
        sampler_kw.update(
            { 'ptform_args' : ( global_dict['handler'].par_prior, ),
              'logl_kwargs' : logl_kw,
              'queue_size' : Ncpu
            }
        ) 
        sampler = Sampler( loglikelihood = loglikelihood,
                           ndim = len( global_dict['handler'].par_free ),
                           sampler = sampler,
                           prior_transform = transform_to_prior_unit_cube,
                           pool = pool,
                           dynesty_sampler_kw = sampler_kw )

        # Run sampling
        sampler.run_sampling( dynesty_sampling_kw = sampling_kw )
    
    if sampler == 'emcee' :
        
        # Build sampler
        sampler_kw.update( { 'kwargs' : logl_kw } )
        sampler = Sampler( loglikelihood = logprob,
                           ndim = len( global_dict['handler'].par_free ),
                           sampler = sampler,
                           nwalkers = nwalkers,
                           pool = pool,
                           emcee_sampler_kw = sampler_kw )

        # Define initial position for walkers
        pos_init = global_dict['handler'].rng.uniform(*global_dict['handler'].par_prior.T,
                                          size=(nwalkers,
                                                len(global_dict['handler'].par_free)))

        # Run sampling
        sampler.run_sampling( pos_init, nsamples,
                              emcee_sampling_kw = sampling_kw )
    
    # return the sampler for storing
    return sampler

################################################################################

def store_results ( sampler, 
                    out_dir = '.', name = '',
                    method = 'hdf5', lightweight = False,
                    pickle_sampler = False, pickle_raw = False ) :
        
    # Store results:
    outbase = generate_output_base( out_dir = out_dir,
                                    name = name )
    _ = dump_results( model = global_dict['model'],
                      handler = global_dict['handler'],
                      data = global_dict['data'],
                      sampler = sampler,
                      noise = global_dict['noise'],
                      outbase = outbase,
                      method = method,
                      lightweight = lightweight
                     )
    sampler.save_results( outbase = outbase,
                          pickle_sampler = pickle_sampler,
                          pickle_raw = pickle_raw )
    return;

################################################################################

def _sample_serial ( which_sampler = 'emcee', nwalkers = None, nsamples = None, 
                     sampler_kw = {}, logl_kw = {}, run_sampling_kw = {}, 
                     out_dir = '.', name = '',
                     store_method = 'hdf5', store_lightweight = False, 
                     pickle_sampler = False, pickle_raw = False  ) :

    # Run the sampling
    sampler = sample(
        sampler = which_sampler, 
        sampler_kw = sampler_kw, 
        logl_kw = logl_kw,
        run_sampling_kw = run_sampling_kw,
        nwalkers = nwalkers,
        nsamples = nsamples
    )
    
    # Store the results
    store_results(
        sampler, 
        out_dir = out_dir,
        name = name,
        method = store_method,
        lightweight = store_lightweight, 
        pickle_sampler = pickle_sampler,
        pickle_raw = pickle_raw
    )
    
    return;

################################################################################

def _sample_parallel ( which_sampler = 'emcee', nwalkers = None, nsamples = None, 
                       sampler_kw = {}, logl_kw = {}, run_sampling_kw = {},
                       Ncpu = None, 
                       out_dir = '.', name = '', 
                       store_method = 'hdf5', store_lightweight = False,
                       pickle_sampler = False, pickle_raw = False  ) :
    import multiprocessing as mp
    
    # Use all available CPUs if no specific number is provided.
    if Ncpu is None :
        Ncpu = mp.cpu_count()

    # Run the sampling
    sampler = None
    with mp.get_context("fork").Pool( Ncpu ) as pool :
        sampler = sample(
            sampler = which_sampler, 
            sampler_kw = sampler_kw, 
            logl_kw = logl_kw,
            run_sampling_kw = run_sampling_kw,
            nwalkers = nwalkers,
            nsamples = nsamples,
            Ncpu = Ncpu, pool = pool
        )

    # Store the results
    store_results(
        sampler, 
        out_dir = out_dir,
        name = name, 
        method = store_method,
        lightweight = store_lightweight, 
        pickle_sampler = pickle_sampler,
        pickle_raw = pickle_raw
    )
    
    return;

################################################################################

def _run () :

    ####################################################################
    # Read command-line arguments:
    
    parser = argparse.ArgumentParser( description = 'options' )
    parser.add_argument( 'parameter_file',
                         default = "",
                         help = 'Path to parameter file')
    parser.add_argument( '-s', '--serial',
                         dest = 'serial',
                         action = 'store_true',
                         help = 'Run the program serially.')
    parser.add_argument( '-mp', '--multiprocessing',
                         dest = 'Ncpu',
                         type = int,
                         default = None,
                         help = ( 'flag for shared-memory parallel run: ' +
                                  'provide number of processes. ' +
                                  'If none is passed defaults to all the available CPUs.') )
    args = parser.parse_args()
    
    spec = importlib.util.spec_from_file_location( "hyper_parameters",
                                                   args.parameter_file )
    hyperpar = importlib.util.module_from_spec( spec )
    spec.loader.exec_module( hyperpar )
    
    ####################################################################

    # Initialization
    init_args = (
        hyperpar.bands,
        hyperpar.fluxes,
        hyperpar.errors,
        hyperpar.uplims if hyperpar.uplims is not None
        else numpy.zeros_like(hyperpar.bands, dtype=bool),
        hyperpar.filters,
        hyperpar.galaxy_parameters,
    )
    init_kwargs = dict(
        sfh_model    = hyperpar.sfh_model,
        ssp_lib      = hyperpar.ssp_lib,
        do_Radio     = hyperpar.do_Radio,
        do_Xray      = hyperpar.do_Xray,
        do_AGN       = hyperpar.do_AGN,
        noise_model  = hyperpar.noise_model,
        noise_params = hyperpar.noise_parameters,
        gxy_kwargs = {
            'lstep' : hyperpar.lstep
                },
        noise_kwargs = hyperpar.noise_kwargs,
        filter_kwargs = hyperpar.filters_custom if hyperpar.filters_custom is not None else {},
    )
    initialize( *init_args, **init_kwargs )

    # Run
    if args.serial :
        _sample_serial(
            which_sampler = hyperpar.sampler,
            nwalkers = hyperpar.nwalkers,
            nsamples = hyperpar.nsamples, 
            sampler_kw = hyperpar.sampler_kw, 
            logl_kw = { 'method_uplims' : hyperpar.method_uplims }, 
            run_sampling_kw = hyperpar.sampling_kw,
            out_dir = hyperpar.output_directory,
            name = hyperpar.run_id,
            store_method = hyperpar.store_method,
            store_lightweight = hyperpar.store_lightweight,
            pickle_sampler = hyperpar.pickle_sampler,
            pickle_raw = hyperpar.pickle_raw
        )
    else :
        _sample_parallel(
            which_sampler = hyperpar.sampler,
            nwalkers = hyperpar.nwalkers,
            nsamples = hyperpar.nsamples, 
            sampler_kw = hyperpar.sampler_kw, 
            logl_kw = { 'method_uplims' : hyperpar.method_uplims }, 
            run_sampling_kw = hyperpar.sampling_kw,
            Ncpu = args.Ncpu, 
            out_dir = hyperpar.output_directory,
            name = hyperpar.run_id,
            store_method = hyperpar.store_method,
            store_lightweight = hyperpar.store_lightweight, 
            pickle_sampler = hyperpar.pickle_sampler,
            pickle_raw = hyperpar.pickle_raw
        )

    return;    

################################################################################
# This commented out here because reasons ...
# if __name__ == '__main__' :    
#     run()
################################################################################

default_parameter_file = """
##################################################
# Parameters for building the observation to fit #
##################################################

# The observed dataset is expressed as 4 array-likes containing respectively:
# - bands: a list of strings corresponding to the unique identifiers also used
#          in the 'filters' variable below 
# - fluxes: measures of fluxes (or upper limits) at the corresponding bands
#           listed in variable 'bands'.
#           The input measure should be given in units of milli-Jansky
# - errors: 1-sigma error on the measures of the fluxes (or upper limits) listed
#           in the variable 'fluxes'
# - uplims: sequence of booleans identifying whether the values listed 
#           in argument ``fluxes`` should be considered non-detection (``True``)
#           or a detection (``False``)
bands  = None
fluxes = None
errors = None
uplims = None

# The photometric filters system used in the observation.
# This parameter should be an iterable containing names
# of filters already present in the database, e.g.
#
# filters = ['GOODS.b', 'GOODS.i', 'GOODS.v', 'GOODS.z']
#
# NOTE that, if the bands listed in variable ``bands`` 
# are all present in the database and have the same names
# of the filters listed with ``galapy.PhotometricSystem.print_filters()``,
# this variable can be set as ``filters = bands``
filters = list(bands)

# Eventual custom photometric filters. 
# This parameter should be a nested dictionary with user-defined transmissions.
# Such transmissions are passed as properly formatted dictionaries
# 
# filters = {{ 'filter_1' : {{ 'wavelengths' : array-like,
#                            'photons' : array-like }},
#             'filter_2' : {{ 'wavelengths' : array-like,
#                            'photons' : array-like }},
#             ...
#             'filter_N' : {{ 'wavelengths' : array-like,
#                            'photons' : array-like }} }}
#
# As the keywords in the lower level dictionaries suggest,
# the two arrays provided, must define the
# wavelength grid and the corresponding transmission in photon units
# (and thus they must have the same size).
# Note that the chosen keyword in the higher level dictionary
# (e.g. 'filter_1') will be used as unique identifier of
# the custom transmission.
filters_custom = None

# Method for treatment of the upper limits, when present.
# Available methods are:
# - 'simple' : a heaviside function which is 0. when the model's flux is
#              smaller than the upper-limit and +infty otherwise
# - 'chi2' : the distance between model and data is expressed as a normal chi-squared
# - 'S12' : modified version of the chi-squared integrating a gaussian, with mean = data-flux
#           and std = data-error, up to the value of the model's flux (Sawicki, 2012)
method_uplims = 'chi2'

########################################
# Parameters defining the galaxy model #
########################################

# Cosmological model to use for computing distances and ages.
# Pre-computed cosmologies present in the database:
# - WMAP7
# - WMAP9
# - Planck15
# - Planck18
# To use another cosmological model the user has to provide instead
# a dictionary with the following key-value couples:
# - 'redshift' : an iterable containing monothonically increasing
#                values of redshift
# - 'luminosity_distance' : an iterable with the luminosity distances
#                           corresponding to the redshift values in
#                           the first iterable
# - 'age' : an iterable with the age of the universe corresponding to
#           the redshift values in the first iterable
cosmo     = 'Planck18'

# The Star-Formation History model to use for building the galaxy model.
# Available models are:
# - 'insitu'
# - 'constant'
# - 'delayedexp'
# - 'lognormal'
sfh_model = '{0:s}'

# The SSP library to use for building the unattenuated stellar emission component
ssp_lib   = 'parsec22.NT'

# Whether to provide X-ray emission support (True) or not (False). 
do_Xray = False

# Whether to provide radio-emission support (True) or not (False).
do_Radio = False

# Whether to build a galaxy containing an AGN (True) or not (False).
do_AGN = False

# Sub-sampling of the wavelength grid.
# If lstep is an integer it will consider a wavelength grid entry every lstep values.
# If lstep is a sequence of integers or a mask, only the wavelength grid entries
# corresponding to the indices provided will be considered.
# If None, it will consider the whole wavelength grid (safest choice)
lstep = None

# Eventual noise model to add. The default is ``None``, i.e. no noise will be added.
# Valid choices for the noise are:
# - 'calibration_error' : accounts for eventual systematics in the calibration of
#                         the errors in the observations.
#
# If this is set to ``None`` all the other hyper-parameters in this file
# related to noise will be ignored.
noise_model = None

# Eventual keyword arguments to be passed to the noise model of choice 
# (leave empty for no keyword arguments)
noise_kwargs = {{}}

#############################################
# Here define the fixed and free parameters #
#############################################

# The dictionary provided here will be used to set the fixed parameters to
# the given value or to flag parameters as 'free' with some associated prior.
# All the parameters that are not used in the specified galaxy model will be ignored.
# (e.g. if the galaxy model has been built with ``do_AGN = False`` all the eventual
#  AGN-related parameters provided will be ignored)
#
# - To set the parameter 'fixed_parameter' as FIXED provide a single value:
#   parameters = {{
#       ...
#       'fixed_parameter' : some_float_value,
#       ...
#   }}
#
# - To set the parameter 'free_parameter' as FREE provide a tuple
#   parameters = {{
#       ...
#       'free_parameter' : ( a_list, a_bool ),
#       ...
#   }}
#   The list ``a_list`` contains the minimum and maximum value of the
#   UNIFORM prior from which to draw samples for the 'free_parameter'.
#   The boolean ``a_bool`` states whether the prior has to be considered
#   * logarithmic: ``a_bool = True``, therefore samples will be drawn from the interval
#                  10**min(a_list) < 'free_parameter' < 10**max(a_list)
#   * linear: ``a_bool = False``, therefore samples will be drawn from the interval
#             min(a_list) < 'free_parameter' < max(a_list)

galaxy_parameters = {{
    
    ##########
    # Galaxy #
    ##########

    'age'      : ( [6., 11.], True ),
    'redshift' : ( [0., 10.], False ),

    ##########################
    # Star Formation History #
    ##########################
    
    'sfh.tau_quench' : ( [6., 11.], True ),
{1:s}
    ########################
    # Inter-Stellar Medium #
    ########################

    'ism.f_MC' : ( [0., 1.], False ),

    ##################
    # Molecular Clouds
    
    'ism.norm_MC' : 100.,
    'ism.N_MC'    : ( [0., 5.], True ),
    'ism.R_MC'    : ( [0., 5.], True ),
    'ism.tau_esc' : ( [4., 8.], True ),
    'ism.dMClow'  : 1.3,
    'ism.dMCupp'  : 1.6,
    
    ##############
    # Diffuse Dust
    
    'ism.norm_DD' : 1.0,
    'ism.Rdust'   : ( [0., 5.], True ),
    'ism.f_PAH'   : ( [0., 1.], False ),
    'ism.dDDlow'  : 0.7,
    'ism.dDDupp'  : 2.0,

    ###############
    # Synchrotron #
    ###############

    'syn.alpha_syn'   : 0.75,
    'syn.nu_self_syn' : 0.2,

    #####################
    # Nebular Free-Free #
    #####################

    'nff.Zgas' : 0.02,
    'nff.Zi'   : 1.,

    ###########################
    # Active Galactic Nucleus #
    ###########################

    'agn.fAGN' : ( [-3., 3.], True ),
    
    'agn.ct' : 40,
    'agn.al' : 0.,
    'agn.be' : -0.5,
    'agn.ta' : 6.,
    'agn.rm' : 60,
    'agn.ia' : 0.001,
    
}}

noise_parameters = {{

    #########
    # Noise #
    #########

    ###################
    # Calibration Error
    
    'f_cal' : ( [-10., 1.], True ),
}}

##############################
# Parameters for the sampler #
##############################

# Choose the sampler, valid options are
# 'emcee' : Affine Invariant MCMC ensemble sampler
# 'dynesty' : Dynamic Nested Sampler
sampler = 'emcee'

# EMCEE SAMPLER-SPECIFIC MANDATORY PARAMETERS
# - set the number of walkers (``nwalkers``)
# - set the chain length (``nsamples``)
nwalkers = 64
nsamples = 4096

# Sampler keyword arguments.
# These are the parameters that will be passed to the constructor of
# the chosen sampler. See relative documentation for further informations:
# - emcee :
# - dynesty :
sampler_kw = {{}}

# Sampling keyword arguments
# These are the parameters that will be passed to the routine running the
# parameter-space sampling. See relative documentation for further informations:
# - emcee :
# - dynesty :
sampling_kw = {{}}

# Output directory (note that if the directory does not exist it will be created)
output_directory = ''

# An identification name, it will be pre-pended to all files stored in the output directory
# (if the string is empty the current date+time will be used)
run_id = ''

# The method used for storing results. 
# Possible choices are:
# - 'hdf5' : uses HDF5 format to save a dictionary containing all the informations
#            to build the Results class used for the analysis and visualization of the
#            sampling-run results. (This is the safest choice, also in terms of security)
# - 'pickle' : uses the standard python pickle format to directly save an instance of
#              the Results class used for the analysis and visualization of the
#              sampling-run results. (Recommended only for local usage, pickled objects
#              are not safe for distribution)
store_method = 'hdf5'

# Only available if the output format chosen is HDF5 (see parameter store_method).
# If True it stores only the chains, weights and loglikelihood values obtained by the
# sampling run (along with information to re-build all the models used in running.
# If False it will save all the infos on the derived quantities as well.
# Selecting lightweight storage, it does not change the way users will ultimatelly access
# the Results object, what will change is the time spent for loading it.
# With lightweight storage the size of the output file is smaller, and the Results object
# is computed (instantiated) when loading the file (so this will take up to some minutes).
# With lightweight storage off, all the quantities are computed at the end of the sampling
# run and are ready to use but the output file can reach a size of up to some GiB. 
store_lightweight = False

# Whether to pickle the sampler raw results.
# (might be useful for analyzing run statistics)
pickle_raw = False

# Whether to pickle the sampler at the end-of-run state.
# (might be useful for extending the run)
pickle_sampler = False

#############################
# ... and that's all folks! #
#############################
"""

def _generate_parameter_file () :

    ####################################################################
    # Read command-line arguments:
    
    parser = argparse.ArgumentParser(
        description = (
            'Writes the parameter file that has to be modified by users '
            'to meet the needs of their sampling.'
        )
    )
    parser.add_argument( '--name', '-n',
                         dest = 'name',
                         type = str,
                         default = 'galapy_hyper_parameters',
                         help = (
                             'provide here the name you want to give to ' +
                             'the parameter file, you can also choose ' +
                             'a path different to the current working directory: ' +
                             '/path/to/chosen_name\n' +
                             'NOTE THAT the extension ".py" will be appended to ' +
                             'the name chosen (this just to guarantee proper formatting ' +
                             'when opening the file in a text editor). ' +
                             'DEFAULT: ${PWD}/galapy_hyper_parameters.py'
                         ) )
    parser.add_argument( '--SFH_model', '-sfh',
                         dest = 'sfh_model',
                         type = str,
                         default = None,
                         help = (
                             'Choose a SFH model. ' +
                             'Available models are:' +
                             ' insitu, constant, delayedexp, lognormal. ' +
                             'DEFAULT: None'
                         ) )
    args = parser.parse_args()
    
    ####################################################################

    sfh_models = set(['insitu', 'constant', 'delayedexp', 'lognormal'])
    sfh_params = {
        'insitu' : """
    # In-Situ model
    'sfh.psi_max' : ( [0., 4.], True ),
    'sfh.tau_star' : ( [6., 11.], True ),
        """,
        'constant' : """
    # Constant model
    'sfh.psi'   : ( [0., 4.], True ),
    'sfh.Mdust' : ( [6., 14.], True ),
    'sfh.Zgxy'  : ( [0., 1.], False ),
        """,
        'delayedexp' : """    
    # Delayed-Exponential model
    'sfh.psi_norm' : ( [0., 4.], True ),
    'sfh.k_shape'  : ( [0., 5.], False ),
    'sfh.tau_star' : ( [6., 11.], True ),
    'sfh.Mdust' : ( [6., 14.], True ),
    'sfh.Zgxy'  : ( [0., 1.], False ),
        """,
        'lognormal' : """
    # Log-Normal model
    'sfh.psi_norm'   : ( [0., 4.], True ),
    'sfh.sigma_star' : ( [0., 5.], False ),
    'sfh.tau_star'   : ( [6., 11.], True ),
    'sfh.Mdust' : ( [6., 14.], True ),
    'sfh.Zgxy'  : ( [0., 1.], False ),
        """,
    }

    sfh_params_string = ''
    if args.sfh_model not in sfh_models and args.sfh_model is not None :
        raise RuntimeError(
            'The chosen model is not available. '
            'To see a list of the available choices call '
            'this function with the --help argument'
        )
    elif args.sfh_model is None :
        args.sfh_model = 'insitu'
        for sfh in sfh_models :
            sfh_params_string += sfh_params[sfh]
    else :
        sfh_params_string = sfh_params[args.sfh_model]

    with open( args.name + '.py', 'w' ) as paramfile :
        paramfile.write( default_parameter_file.format(args.sfh_model, sfh_params_string) )

    return;

################################################################################



# External imports
import numpy
import os
from collections.abc import MutableMapping as MM

# Internal imports
import galapy.internal.globs as GP_GBL
from galapy.internal.data import DataFile
from galapy.internal.utils import FlagVal
from galapy.BandpassTransmission import BPT

def _recursive_print_filters ( root, pathline = [] ) :
    _, subd, files = next(os.walk(root))
    for f in files : print('.'.join(pathline+f.split('.')[:-1]))
    for s in subd : _recursive_print_filters(os.path.join(root, s), pathline+[s])
    return;    

def print_filters ( experiment = None ) :
    """Prints on screen all filters available in database.
    
    Parameters
    ----------
    experiment : string or None
        (Optional, default = None) filter the results by experiment.
    
    Examples
    --------
    >>> print_filters(experiment='Herschel.SPIRE')
    
    # Path: /path/to/.galapy
    
    Herschel.SPIRE.PSW
    Herschel.SPIRE.PLW
    Herschel.SPIRE.PMW
    """
    from galapy.configuration import rcParams
    for path in rcParams['datapath'] :
        print(f"\n# Path: {path}\n")
        if experiment is None or len(experiment) == 0 :
            _recursive_print_filters( root = os.path.join(path, *GP_GBL.FILT_DIR ),
                                      pathline = [] )
        else :
            try :
                exppath = experiment.split('.')
                _recursive_print_filters( root = os.path.join(
                    path, *GP_GBL.FILT_DIR, *experiment.split('.') ),
                                          pathline = exppath )
            except :
                raise AttributeError( f'experiment {experiment} not in database.' )
    return;

def _recursive_list_filters ( root, pathline = [], outlist = [] ) :
    _, subd, files = next(os.walk(root))
    for f in files : outlist += ['.'.join(pathline+f.split('.')[:-1])]
    for s in subd : _recursive_list_filters(os.path.join(root, s), pathline+[s], outlist)
    return;    

def list_filters ( experiment = None ) :
    """Return a list of filters available in database.
    
    Parameters
    ----------
    experiment : string or None
        (Optional, default = None) filter the results by experiment.

    Returns
    -------
    : list
    a list of strings, containing the name of each filter in the database
    
    Examples
    --------
    >>> list_filters(experiment='Herschel.SPIRE')
    ['Herschel.SPIRE.PSW', 'Herschel.SPIRE.PLW', 'Herschel.SPIRE.PMW' ]
    """
    from galapy.configuration import rcParams
    outlist = []
    for path in rcParams['datapath'] :
        if experiment is None or len(experiment) == 0 :
            _recursive_list_filters( root = os.path.join(path, *GP_GBL.FILT_DIR ),
                                     pathline = [], outlist = outlist )
        else :
            try :
                exppath = experiment.split('.')
                _recursive_list_filters(
                    root = os.path.join(
                        path, *GP_GBL.FILT_DIR, *experiment.split('.') ),
                    pathline = exppath, outlist = outlist )
            except :
                raise AttributeError( f'experiment {experiment} not in database.' )
    return outlist

class PMS () :
    """ Builds the Photometric System 
    
    Parameters
    ----------
    *args : tuple of strings
       As many strings as necessary. Each string has to name
       one of the available filters already present in the database.
    **kwargs : dictionary
       Keyword arguments can be used for providing user-defined 
       transmissions. 
       Custom transmissions are passed as properly formatted dictionaries:
       ``keyword = { 'wavelengths' : array-like,'photons' : array-like }``
       The two arrays provided, as the keys suggest, must define the
       wavelength grid and the corresponding transmission in photon units
       (and thus they must have the same size).
       Note that the chosen keyword will be used as unique identifier of
       the custom transmission.
    """

    def __init__ ( self, *args, **kwargs ) :

        # Build the Bandpass-Transmissions dictionary:
        self.bpt = {}

        # allocate an array to append filter limits
        extremes = numpy.array([], dtype=object)
        
        # From positional arguments pass band-names
        if len( args ) > 0 :
            for arg in args :
                try :
                    *path, name = arg.split('.')
                    ll, fl = numpy.loadtxt( DataFile(
                        f'{name:s}.dat',
                        numpy.append( GP_GBL.FILT_DIR, path )
                    ).get_file(), unpack=True )
                except OSError :
                    raise ValueError(
                        f'Filter "{arg}" provided is not present in galapy database.'
                    )
                except :
                    raise ValueError(
                        f'Argument "{arg}" provided does not name a known filter.'
                    )

                idx_sort = numpy.argsort( ll )
                ll = numpy.ascontiguousarray(ll[idx_sort])
                fl = numpy.ascontiguousarray(fl[idx_sort])

                try :
                    self.bpt[ arg ] = BPT( ll, fl )
                except :
                    ValueError()

                # append interval extremes to flattened array of flagged values
                # ('s' = start, 'e' = end)
                extremes = numpy.append( FlagVal( self.bpt[arg].get_lmin(), 's' ), extremes )
                extremes = numpy.append( FlagVal( self.bpt[arg].get_lmax(), 'e' ), extremes )

        # From keyword arguments pass custom bands
        for key, value in kwargs.items() :
            if isinstance( value, MM ) :
                try :
                    idx_sort = numpy.argsort( value[ 'wavelengths' ] )
                    ll = numpy.ascontiguousarray(value[ 'wavelengths' ][idx_sort])
                    fl = numpy.ascontiguousarray(value[ 'photons' ][idx_sort])
                    self.bpt[ key ] = BPT( ll, fl )
                except KeyError :
                    raise ValueError(
                        f'The provided dictionary "{key}" does not contain the necessary items, '
                        'please provide a "wavelengths" and a "photons" array.'
                    )
                except :
                    raise ValueError()
                
                # append interval limits to flattened array of flagged values
                # ('s' = start, 'e' = end)
                extremes = numpy.append( FlagVal( self.bpt[key].get_lmin(), 's' ), extremes )
                extremes = numpy.append( FlagVal( self.bpt[key].get_lmax(), 'e' ), extremes )
                
            else :
                raise ValueError( 'Keyword arguments must be correctly formatted dictionaries: '
                                  'k = {"wavelengths":array-like,"photons":array-like}' )

        # Sort the flattened extremes array
        extremes = extremes[numpy.argsort([ext.value for ext in extremes])]

        # remove overlappings in O(NlogN) time (with N = number of intervals):
        self.olims = []
        count = 0
        for ext in extremes :
            if ext.flag == 's' :
                if count == 0 :
                    self.olims += [[ext.value]]
                count += 1 
            if ext.flag == 'e' :
                count -= 1
                if count == 0 :
                    self.olims[-1] += [ext.value]
        self.olims = numpy.asarray( self.olims )

        # Store array with pivot wavelengths
        self.lpiv = numpy.asarray( [ bpt.get_lpiv()
                                     for bpt
                                     in self.bpt.values() ] )
        self.keys = numpy.asarray( [ k for k in self.bpt.keys() ] )
        
        idxsort = numpy.argsort( self.lpiv )
        self.lpiv = self.lpiv[idxsort]
        self.keys = tuple( self.keys[idxsort] )

    def dump ( self ) :
        return {
            k : { 'wavelengths' : numpy.asarray( v.get_xaxis() ),
                  'photons'     : (
                      numpy.asarray( v.get_yaxis() ) *
                      numpy.asarray( v.get_xaxis() ) /
                      v.get_norm() ) }
            for k, v in self.bpt.items()
        }

    @classmethod
    def load ( cls, dictionary ) :
        return cls( **dictionary )

    def __len__ ( self ) :
        return len( self.bpt )

    def get_intervals ( self ) :
        """
        """
        pass
    
    def get_fluxes ( self, ll, fl ) :
        """ Get a list of fluxes, one per each band in the photometric system.
        
        Parameters
        ----------
        ll : array-like
        
        fl : array-like
        
        Returns
        -------
        : list
        """
        fluxes = []
        for key in self.keys :
            wl, wu = [ numpy.argmax( self.bpt[key].get_lmin() < ll ),
                       numpy.argmin( ll < self.bpt[key].get_lmax() ) ]
            fluxes += [self.bpt[key].get_bandpass_flux( ll[wl:wu], fl[wl:wu] )]
    
        return fluxes




########################################################################################

# External imports
import numpy
import warnings

# Internal imports
from galapy.Galaxy import GXY
from galapy.Noise import Noise
from galapy.internal.abc import Model
from galapy.internal.utils import set_nested, unwrap_items

########################################################################################

class ModelParameters () :
    """A class to handle model parameters for GalaPy objects of type Model.

    This class provides methods for setting up and handling model parameters,
    both fixed and free, for galaxy and noise models.

    Parameters
    ----------
    *args : galapy.internal.abc.Model
        The galapy.Galaxy.GXY and/or galapy.Noise.Noise models to extract parameters from.
    sample_params : dict, optional
        A dictionary containing fixed and free parameter values for sampling.
        The keys are parameter names, and the values are either fixed values or tuples
        containing (prior_distribution, log_prior), where prior_distribution is a 
        2-elements iterables defining the limits of the uniform prior and log_prior
        is a boolean defining whether the prior has to be considered logarithmic (base 10) 
        or not.
    rng_kw : dict, optional
        Keyword arguments to be passed to numpy.random.default_rng().

    Attributes
    ----------
    parameters : dict
        A dictionary containing all model parameters.
    par_free : numpy.ndarray
        An array containing names of free parameters.
    par_log : numpy.ndarray
        An array containing booleans defining which of 
        the free parameters are treated logarithmically.
    par_prior : numpy.ndarray
        An array containing prior distributions for free parameters.

    Examples
    --------
    >>> from galapy.Galaxy import GXY
    >>> from galapy.Handlers import ModelParameters

    Instantiate a galaxy model

    >>> gxy = GXY()

    Define sample parameters

    >>> sample_params = {'galaxy.age': 1.e+8, 'galaxy.sfh.psi_max': ([0., 5.], True)}

    Create ModelParameters instance

    >>> params = ModelParameters(gxy, sample_params)

    Return nested dictionary of parameters

    >>> nested = params.return_nested()
    >>> gxy.set_parameters( **nested['galaxy'] )

    Dump instance to a dictionary

    >>> dumped_params = handle.dump()
    
    Build instance from dumped dictionary

    >>> handle2 = ModelParameters.load(dumped_params)
    
    the 2 handlers are now equivalent:

    >>> handle.return_nested() == handle2.return_nested()
    True
    
    but they are not the same object

    >>> handle is handle2
    False
    """

    def __init__ ( self, *args, sample_params = {}, rng_kw = {} ) :

        # Set the default random engine 
        self.rng = numpy.random.default_rng(**rng_kw)
        
        # Extract parameterisation from the models
        self.parameters = {}
        _i = 0
        for model in args :
            if not isinstance( model, Model ) :
                raise RuntimeError(
                    f'The provided model ``{model}`` is not valid. '
                    'Valid models are instances of class galapy.internal.abc.Model'
                )
            if isinstance( model, GXY ) :
                model_key = ['galaxy']
            elif isinstance( model, Noise ) :
                model_key = ['noise'] 
            else :
                model_key = [f'custom{_i}'] 
                _i += 1
            self.parameters.update(
                { '.'.join(model_key + k) : v for k, v in unwrap_items(model.params) }
            )
            
        # Read fixed and free parameters from dictionary
        self.par_log   = []
        self.par_free  = []
        self.par_prior = []
        for key, value in sample_params.items() :
            if key not in self.parameters.keys() :
                warnings.warn(
                    f'Parameter "{key}" is not present in the model '
                    'requested and will be ignored.')
                continue
            if isinstance(value, tuple) :
                prior, log = value
                self.par_free.append( key )
                self.par_prior.append( prior )
                self.par_log.append( log )
                v = self.rng.uniform( *prior, size=None )
                self.parameters[key] = 10.**v if log else v
            else :
                try :
                    self.parameters[key] = float(value)
                except ValueError as ve :
                    warnings.warn( f'The value ({value}) provided for the fixed parameter {key} '
                                   'is not valid and will be ignored. '
                                   f'Using the default value {self.parameters[key]} instead.' )
        self.par_log   = numpy.asarray(self.par_log)
        self.par_free  = numpy.asarray(self.par_free)
        self.par_prior = numpy.asarray(self.par_prior)

    @classmethod
    def load ( cls, dictionary ) :
        """Build a ModelParameters instance loading from a dictionary.

        Parameters
        ----------
        dictionary : dict
            A formatted dictionary containing stored info from which to define the 
            new ModelParameters instance.

        Returns
        -------
        ModelParameters
            A ModelParameters instance with loaded parameters.

        Raises
        ------
        KeyError
            If an element of the input dictionary is not an attribute of the class.
        """
        ret = cls()
        for k, v in dictionary.items() :
            if k not in ret.__dict__.keys() :
                raise KeyError(
                    f'Element {k} of input dictionary not an attribute of the class'
                )
            if k == 'par_free' :
                v = numpy.asarray(v.split('|'))
            setattr( ret, k, v )
        return ret
        
    def dump ( self ) :
        """Dump a ModelParameters instance to a dictionary.

        Returns
        -------
        dict
            A dictionary containing dumped attributes of the current instance.
        """
        return dict(
            parameters = self.parameters,
            par_free   = '|'.join(self.par_free),
            par_log    = self.par_log,
            par_prior  = self.par_prior,
        )
                    
    def return_nested ( self, par = None ) :
        """ From a list of parameters returns a nested dictionary in 
        the proper format. Here 'proper' means that such dictionary has
        the hierarchy necessary to pass it as keyword arguments dictionary
        to an object of type galapy.internal.abc.Model()
        
        Parameters
        ----------
        par : array-like or iterable
          a list of values to be assigned to the parameters flagged as `free`
        
        Returns
        -------
        : dict
          A nested dictionary with the proper format to be passed as keyword 
          arguments to the constructor or to function `set_parameters` of 
          objects of type galapy.internal.abc.Model().
        
        Examples
        --------
        Given a galaxy model of type GXY:

        >>> from galapy.Galaxy import GXY
        >>> gxy = GXY()

        and a dictionary with the fixed and free parameters as follows
        
        >>> p = { 'age' : 1.e+8, 'sfh.psi_max' : ( [0., 5.], True ) }

        we can build a parameter handler:
 
        >>> from galapy.GalaxyParameters import GXYParameters
        >>> par = GXYParameters( gxy, p )
        >>> par.par_free
        numpy.array( [ 'sfh.psi_max' ] )
        
        Now, assuming we want to set a new value to the 'sfh.psi_max' parameter,

        >>> nested = par.return_nested( [ 123. ] )
        >>> nested 
        { 'sfh' : { 'psi_max' : 123. } }
        
        This dictionary can be used to set the parameters of the galaxy model

        >>> gxy.set_parameters( **nested )
        """

        # If no argument is passed use internal parameters
        ret = {}
        if par is None :
            ret = {}
            for klist, v in self.parameters.items() :
                set_nested( ret, klist.split('.'), v )
            return ret
        
        # Otherwise use free parameters
        par = numpy.asarray( par ) 
        if len(par) != len(self.par_free) :
            raise RuntimeError( f'Provided {len(par)} but there are exactly '
                                f'{len(self.par_free)} free parameters to set.' )
        val = numpy.zeros_like(par)
        val[self.par_log] = 10.**par[self.par_log]
        val[~self.par_log] = par[~self.par_log]
        for v, klist in zip( val, self.par_free ) :
            set_nested( ret, klist.split('.'), v )
        return ret


########################################################################################

class GXYParameters ( ModelParameters ) :
    """A class to handle model parameters specifically for galapy.galaxie.GXY() type objects.

    This class inherits from ModelParameters and provides methods tailored for galaxy models.

    Parameters
    ----------
    gxy_model : galapy.Galaxy.GXY
        The galaxy model to extract parameters from.
    sample_params : dict
        A dictionary containing fixed and free parameter values for sampling.
        The keys are parameter names, and the values are either fixed values or tuples
        containing (prior_distribution, log_prior).
    rng_kw : dict, optional
        Keyword arguments to be passed to numpy.random.default_rng().
    """
    def __init__ ( self, gxy_model, sample_params, rng_kw = {} ) :

        super().__init__( gxy_model,
                          sample_params = { '.'.join(['galaxy',k]) : v
                                            for k,v in sample_params.items() },
                          rng_kw = rng_kw )

    def return_nested ( self, par ) :
        """Return model parameters as a nested dictionary specifically for galaxies.

        Parameters
        ----------
        par : array-like or iterable
            A list of values to be assigned to free parameters.

        Returns
        -------
        dict
            A nested dictionary with galaxy model parameters.
        """
        return super().return_nested( par )['galaxy']
    

########################################################################################

class NoiseParameters ( ModelParameters ) :
    """A class to handle model parameters specifically for galapy.Noise.Noise() type objects.

    This class inherits from ModelParameters and provides methods tailored for noise models.

    Parameters
    ----------
    noise_model : galapy.Noise.Noise
        The noise model to extract parameters from.
    sample_params : dict
        A dictionary containing fixed and free parameter values for sampling.
        The keys are parameter names, and the values are either fixed values or tuples
        containing (prior_distribution, log_prior).
    rng_kw : dict, optional
        Keyword arguments to be passed to numpy.random.default_rng().
    """
    def __init__ ( self, noise_model, sample_params, rng_kw = {} ) :

        super().__init__( noise_model,
                          sample_params = { '.'.join(['noise',k]) : v
                                            for k,v in sample_params.items() },
                          rng_kw = rng_kw )

    def return_nested ( self, par ) :
        """Return model parameters as a nested dictionary specifically for noise.

        Parameters
        ----------
        par : array-like or iterable
            A list of values to be assigned to free parameters.

        Returns
        -------
        dict
            A nested dictionary with noise model parameters.
        """
        return super().return_nested( par )['noise']

########################################################################################



""" The Synchrotron module implements a generic parameterized synchrotron emission.
"""

# External imports
import numpy

# Internal imports
from galapy.internal.constants import clight
from .SYN_core import CSYN

__all__ = [ 'SYN', 'SNSYN' ]

syn_tunables = ( 'alpha_syn', 'nu_self_syn' )

def syn_build_params ( **kwargs ) :
    """ Builds the parameters dictionary for the generic synchrotron.
    """
    
    out = {
        'alpha_syn'   : 0.75,
        'nu_self_syn' : 0.2,  # [GHz]
    }
    for k in set( out.keys() ).intersection(kwargs.keys()) :
        out[k] = kwargs[k]
        
    return out

class SYN () :
    """ Class wrapping the C-core implementation of a generic Synchrotron emission type.   
    
    Parameters
    ----------
    ll : float array
      the wavelength grid where the synchrotron emission is computed
    """

    def __init__ ( self, ll, **kwargs ) :

        self.l = numpy.ascontiguousarray( ll )
        self.AngUnit = clight['A/s'] / self.l**2
        self.core = CSYN( self.l )
        self.params = syn_build_params( **kwargs )
        self.set_parameters()
        
    def set_parameters ( self, **kwargs ) :
        r""" Function for setting the parameters of the model.
        
        Returns
        -------
        : None
        """

        self.params.update( kwargs )
        self.core.set_params( numpy.asarray( [
            self.params[k]
            for k in syn_tunables
        ], dtype=float) )
        return;

    def opt_depth ( self, il = None ) :        

        if il is None :
            il = numpy.arange( self.l.size, dtype = numpy.uint64 )
        il = numpy.asarray( il )
        scalar_input = False
        if il.ndim == 0 :
            il = il[None] # makes il 1D
            scalar_input = True

        ret = numpy.array( [ self.core.opt_depth( _i ) for _i in il ] )
        if scalar_input :
            return ret.item()
        return ret

    def energy ( self, SynNorm, il = None ) :
        """ Returns the normalized total energy radiated at given wavelength.       
        """
        
        if il is None :
            il = numpy.arange( len(self.l), dtype = numpy.uint64 )
        il = numpy.asarray( il )
        scalar_input = False
        if il.ndim == 0 :
            il = il[None] # makes il 1D
            scalar_input = True
        
        ret = self.core.energy( il, SynNorm )
        if scalar_input :
            return ret.item()
        return ret

    def emission ( self, SynNorm, il = None ) :
        
        return self.energy( SynNorm, il ) * self.AngUnit
    
class SNSYN ( SYN ) :

    NormFact = 1.e+30 # [ erg s^-1 Hz^-1 ]
    Lsyn0    = 3.e+28 # [ erg s^-1 Hz^-1 ]
    eta      = 2.     # [ adimensional ] 

    def __init__ ( self, *args, **kwargs ) :

        super().__init__( *args, **kwargs )
        self.params['RCCSN'] = 0.
        self.params.update(kwargs)

    def emission ( self, il = None ) :
        
        Lsyn = super().energy( self.params['RCCSN'] * SNSYN.NormFact, il )
        wn0 = Lsyn > 0.
        Lsyn[wn0] /= 1 + ( SNSYN.Lsyn0 / Lsyn[wn0] )**SNSYN.eta

        if il is None :
            return Lsyn * self.AngUnit
        return Lsyn * self.AngUnit[ il ]



""" Definitions of physical and mathematical constants
"""

# Solar luminosity [ erg / s ]
Lsun = 3.828e+33
sunL = 1. / Lsun

# Speed of light in vacuum
clight = {
    'cm/s' : 2.99792458e+10,
    'm/s'  : 2.99792458e+8,
    'A/s'  : 2.99792458e+18,
}

# Mpc to cm
Mpc_to_cm = 3.086e+24

# Planck constant
hP = {
    'eV/Hz' : 4.1357e-15,
    'erg*s' : 6.6262e-27,
}

# Angstrom to keV
def Ang_to_keV ( wavelength ) :
    return 1.e-3 * hP['eV/Hz'] * clight['A/s'] / wavelength




""" Implements the class used for storing sampling results.
"""

#############################################################################################
# External imports

import warnings
import numpy
import pickle
from collections.abc import MutableMapping as MM

#############################################################################################
# Internal imports

import galapy
from galapy.Galaxy import GXY, PhotoGXY
from galapy.Noise import Noise, CalibrationError
from galapy.Handlers import ModelParameters, GXYParameters, NoiseParameters
from galapy.sampling.Observation import Observation
from galapy.sampling.Sampler import Sampler
from galapy.internal.utils import now_string, func_scalar_or_array, quantile_weighted, get_credible_interval, find_nearest
from galapy.io.hdf5 import write_to_hdf5, load_from_hdf5

#############################################################################################

def generate_output_base ( out_dir = '', name = '' ) :
    """
        
    Parameters
    ----------
    out_dir : string
        Position in the filesystem where the results will be stored.
        Default to the directory where the command has been called.
    name : string 
        A string identifying the run that will be saved. 
        By default it will use a string with the current date+time 
    """
    import os
    
    # If no output directory is passed, set it to the current working directory
    if len( out_dir ) == 0 or out_dir is None :
        out_dir = os.getcwd()
        
    # First check whether the required output directory exists,
    # if not it will be created in the correct position of the file-system
    if not os.path.isdir( out_dir ) :
        try :
            # creates multi-level subdirs. similarly to the *Nix command `mkdir -p`
            # (while os.mkdir() only allows to create the highest level directory)
            os.makedirs( out_dir ) 
        except OSError:
            print ( f"Creation of the directory {out_dir} failed" )
            
    # If no name for the current run has been provided, set it to
    # a string with the current date+time: 'year+month+day+hour+minute'
    if len(name) == 0 :
        name = now_string()

    # Set the string with the output base name
    outbase = os.path.join( out_dir, name )

    return outbase

def dump_results ( model, handler, data, sampler,
                   noise = None, outbase = '',
                   method = 'hdf5', lightweight = False ) :
    from time import time

    if len(outbase) == 0 :
        outbase = nowstring()

    if not isinstance( sampler, Sampler ) :
        raise ValueError( "Argument ``sampler`` should be an instance of type Sampler" )
    sample_res, sample_logl, sample_weights = sampler.return_samples_logl_weights()

    if not isinstance( model, GXY ) :
        raise ValueError( "Argument ``model`` should be an instance of type GXY" )

    if not isinstance( handler, ModelParameters ) :
        raise ValueError( "Argument ``handler`` should be an instance of type ModelParameters" )

    if not isinstance( data, Observation ) :
        raise ValueError( "Argument ``data`` should be an instance of type Observation" )

    if noise is not None :
        if not isinstance( noise, Noise ) :
            raise ValueError( "Argument ``noise`` should be an instance of type Noise" )

    if lightweight :
        if method in { 'hdf5', 'h5' } :
            outfile = '_'.join( [ outbase, sampler.which_sampler, 'results_light.galapy.hdf5' ] )
            write_to_hdf5(
                outfile,
                metadata = dict(
                    storage_method = 'light',
                    galapy_version = galapy.__version__,
                ),
                hard = True,
                results = {
                    'model' : model.dump(),
                    'handler' : handler.dump(),
                    'sample_res' : sample_res,
                    'sample_logl' : sample_logl,
                    'sample_weights' : sample_weights,
                    'data' : None if data is None else data.dump(),
                    'noise' : None if noise is None else noise.dump(),
                    'sampler_name' : sampler.which_sampler,
                }
            )
            return outbase
        else :
            warnings.warn(
                "Lightweight (`lightweight=True`) dumping only available with `method = 'hdf5'`. "
                "Falling back to HDF5 output."
            )
                
    print( 'Now processing the sampling results, this might require some time ...' )
    tstart = time()
    results = Results( model, handler,
                       sample_res, sample_logl,
                       sample_weights = sample_weights,
                       data = data, noise = noise,
                       sampler_name = sampler.which_sampler )
    ndur = time() - tstart
    print( f'... done in {ndur} seconds.' )

    if method == 'pickle' :
        # Pickle the Results instance
        with open( '_'.join( [ outbase, sampler.which_sampler,
                               'results.galapy.pickle' ] ), 'wb' ) as pfw :
            pickle.dump( results, pfw )
    if method in { 'hdf5', 'h5' } :
        outfile = '_'.join( [ outbase, sampler.which_sampler, 'results.galapy.hdf5' ] )
        write_to_hdf5(
            outfile,
            metadata = dict(
                storage_method = 'heavy',
                galapy_version = galapy.__version__,
            ),
            hard = True,
            results = results.dump()
        )

    print( f'Results stored in files with prefix: {outbase}' )

    return outbase

def load_results ( infile, method = None, lightweight = None ) :

    if method is None :
        method = infile.split('.')[-1]

    res = None
    
    if method in { 'h5', 'hdf5' } :
        res_dict = load_from_hdf5( infile )
        if lightweight is None :
            lightweight = res_dict['metadata']['storage_method'] == 'light'
        if lightweight :
            from time import time
            print( 'Now processing the sampling results, this might require some time ...' )
            tstart = time()
            res = Results(
                model = (
                    PhotoGXY.load(res_dict['results']['model'])
                    if 'pms_kwargs' in res_dict['results']['model']
                    else GXY.load(res_dict['results']['model'])
                ),
                handler = ModelParameters.load( res_dict['results']['handler'] ),
                sample_res = res_dict['results']['sample_res'],
                sample_logl = res_dict['results']['sample_logl'],
                sample_weights = (
                    res_dict['results']['sample_weights']
                    if 'sample_weights' in res_dict['results']
                    else None
                ),
                data = (
                    Observation.load( res_dict['results']['data'] )
                    if 'data' in res_dict['results']
                    else None
                ),
                noise = (
                    CalibrationError.load( res_dict['results']['noise'] )
                    if res_dict['results']['noise'] is not None
                    else None
                ),
                sampler_name = res_dict['results']['sampler_name']
            )
            ndur = time() - tstart
            print( f'... done in {ndur} seconds.' )
        else :
            res = Results.load( res_dict['results'] )
        return res

    if method == 'pickle' :
        with open( infile, 'rb' ) as pfr :
            res = pickle.load( pfr )
        return res

#############################################################################################

class Results () :
    def __init__ ( self, model, handler, sample_res, sample_logl,
                   sample_weights = None, data = None, noise = None,
                   sampler_name = 'dynesty' ) :
        """ A class for storing the results of a sampling run.
        
        Parameters
        ----------
        model : galapy.Galaxy.GXY
            An instance of type GXY (or derived). It stores the model's architecture
            used for running sampling.
        handler : galapy.Handlers.ModelParameters 
            An instance ot type ModelParameters with the parameterisation used in
            the sampling run.
        sample_res : numpy.ndarray
            the matrix containing all the samples of the run
        sample_logl : numpy.ndarray
            1D array with the loglikelihood values corresponding to the samples
        sample_weights : numpy.ndarray
            (Optional) 1D array with the weight of each sample in the run. 
            Default is ``None``, in which case the array will be padded with ones
            (i.e. all samples have the same weight)
        data : galapy.sampling.Observation.Observation
            (Optional) An instance of type ``Observation`` with the fluxes 
            measurements used in the sampling run
        noise : galapy.Noise.Noise
            (Optional) An instance of type ``Noise`` with
            the eventual noise model used in the sampling run
        sampler_name : str
            Which sampler has been used in the sampling run (i.e. 'dynesty' or 'emcee', 
            default is 'dynesty')
        """
        
        # Store the model architecture
        if not isinstance(model, GXY) and not isinstance( model, MM ) :
            raise AttributeError( 
                'Attribute "model" should be an instance of type ``GXY``'
            )
        if isinstance( model, MM ) :
            self._mod = model
            model = self.get_model()
        else :
            self._mod = model.dump()
        
        # Store the parameters specs
        if not isinstance(handler, ModelParameters) and not isinstance( handler, MM ) :
            raise AttributeError( 
                'Attribute "handler" should be an instance of type ``ModelParameters``'
            )
        if isinstance( handler, MM ) :
            self._han = handler
            handler = self.get_handler()
        else :
            self._han = handler.dump()

        if sample_weights is None :
            sample_weights = numpy.ones_like( sample_logl )
        
        # Store the observation
        self.Ndof = 1
        self._obs = None
        if data is not None :
            if not isinstance(data,Observation) and not isinstance( data, MM ) :
                raise AttributeError( 
                    'Attribute "data" should be an instance of type ``Observation``'
                )
            if isinstance( data, MM ) :
                self._obs = data
                data = self.get_observation()
            else :
                self._obs = data.dump()
            self.Ndof = len(data.pms) - len(handler.par_free)

        # Store the noise model    
        self._noise = None
        if noise is not None :
            if not isinstance( noise, Noise ) and not isinstance( noise, MM ) :
                raise AttributeError( 
                    'Attribute "noise" should be an instance of type ``Noise``'
                )
            if isinstance( noise, MM ) :
                self._noise = noise
                noise = self.get_noise()
            else :
                self._noise = noise.dump()

        # Store the sampler's name and specs
        self.sampler = sampler_name
        self.ndim = len( handler.par_free )
        
        self.size = len(sample_res)
        if self.size != len(sample_logl) or self.size != len(sample_weights):
            raise RuntimeError( 
                'Arguments sample_res, sample_logl and sample_weights should have same length'
            )
        self.params = []
        self.logl    = numpy.asarray( sample_logl )
        self.samples = numpy.asarray( sample_res )
        self.weights = numpy.asarray( sample_weights )
        self.wnot0 = ( self.weights > 0. )
        self.SED    = numpy.empty(shape=(self.size, 
                                         *model.wl().shape))
        self.Mstar  = numpy.empty(shape=(self.size,))
        self.Mdust  = numpy.empty(shape=(self.size,))
        self.Mgas   = numpy.empty(shape=(self.size,))
        self.Zstar  = numpy.empty(shape=(self.size,))
        self.Zgas   = numpy.empty(shape=(self.size,))
        self.SFR    = numpy.empty(shape=(self.size,))
        self.TMC    = numpy.empty(shape=(self.size,))
        self.TDD    = numpy.empty(shape=(self.size,))

        for i, par in enumerate(sample_res) :
            self.params += [handler.return_nested(par)['galaxy']]

            try :
                model.set_parameters( **self.params[-1] )
            except RuntimeError :
                self.SED[i]   = -numpy.inf * numpy.ones_like(model.wl())
                self.Mstar[i] = -numpy.inf
                self.Mdust[i] = -numpy.inf
                self.Mgas[i]  = -numpy.inf
                self.Zstar[i] = -numpy.inf
                self.Zgas[i]  = -numpy.inf
                self.SFR[i]   = -numpy.inf
                self.TMC[i]   = -numpy.inf
                self.TDD[i]   = -numpy.inf
                continue
            
            age = model.age
            self.SED[i]   = model.get_SED()
            self.Mstar[i] = model.sfh.Mstar(age)
            self.Mdust[i] = model.sfh.Mdust(age)
            self.Mgas[i]  = model.sfh.Mgas(age)
            self.Zstar[i] = model.sfh.Zstar(age)
            self.Zgas[i]  = model.sfh.Zgas(age)
            self.SFR[i]   = model.sfh(age)
            self.TMC[i]   = model.ism.mc.T
            self.TDD[i]   = model.ism.dd.T

    def dump ( self ) :
        return dict(
            # Models' architecture
            model = self._mod,
            handler = self._han,
            data = self._obs,
            noise = self._noise,
            # Sampling run hyperparameters
            sampler_name = self.sampler,
            size = self.size, Ndof = self.Ndof,
            # Sampling run stored quantities
            logl    = self.logl,
            samples = self.samples,
            weights = self.weights,
            wnot0   = self.wnot0,
            # Derived quantities
            SED     = self.SED,
            Mstar   = self.Mstar,
            Mdust   = self.Mdust,
            Mgas    = self.Mgas,
            Zstar   = self.Zstar,
            Zgas    = self.Zgas,
            SFR     = self.SFR,
            TMC     = self.TMC,
            TDD     = self.TDD,
        )

    @classmethod
    def load ( cls, dictionary ) :

        # build object
        ret = cls(
            model   = dict( dictionary['model'] ),
            handler = dict( dictionary['handler'] ),
            data    = dict( dictionary['data'] ),
            noise   = dict( dictionary['noise'] ),
            sampler_name = dictionary['sampler_name'],
            sample_res     = [],
            sample_logl    = [],
            sample_weights = []
        )
        # Sampling run stored quantities
        ret.logl    = dictionary['logl']
        ret.samples = dictionary['samples']
        ret.weights = dictionary['weights']
        ret.wnot0   = dictionary['wnot0']

        # Additional hyperparameters
        ret.Ndof    = dictionary['Ndof']
        ret.size    = dictionary['size']

        # Derived quantities
        ret.SED     = dictionary['SED']
        ret.Mstar   = dictionary['Mstar']
        ret.Mdust   = dictionary['Mdust']
        ret.Mgas    = dictionary['Mgas']
        ret.Zstar   = dictionary['Zstar']
        ret.Zgas    = dictionary['Zgas']
        ret.SFR     = dictionary['SFR']
        ret.TMC     = dictionary['TMC']
        ret.TDD     = dictionary['TDD']

        # Compute parameters' dictionaries
        handler = ret.get_handler()
        for i, par in enumerate(ret.samples) :
            ret.params += [handler.return_nested(par)['galaxy']]

        return ret
        
    def get_stored_quantities ( self ) :
        """ Returns a list with all the quantities stored in the instance
        """
        return list( self.__dict__.keys() )

    def get_residuals ( self, which_model = 'bestfit', standardised = True ) :
        """
        """
        _gxy = self.get_model() 
        _obs = self.get_observation()
        _han = self.get_handler()
        if which_model == 'bestfit' :
            params = self.get_bestfit( 'samples' )
        elif which_model == 'mean' :
            params = self.get_mean( 'samples' )
        elif which_model == 'median' :
            params = self.get_quantile( 'samples' )
        else :
            warnings.warn( 'Choice invalid, falling back to default choice (="bestfit")' )
            params = self.get_bestfit( 'samples' )
            
        nested = _han.return_nested(params)
        _gxy.set_parameters( **nested['galaxy'] )
        
        model = _gxy.photoSED()
        data  = _obs.fluxes
        if standardised :
            error = _obs.errors
            if hasattr(self, '_noise') and self._noise is not None :
                _noi = self.get_noise()
                _noi.set_parameters( **nested['noise'] )
                error = _noi.apply( error, model )
            return ( data - model ) / error
        return data - model

    def get_chi2 ( self, which_model = 'bestfit', reduced = True ) :
        """
        """
        chi = self.get_residuals( which_model, standardised = True )
        redfact = 1.
        if reduced :
            redfact /= self.Ndof
        return numpy.sum( chi**2 ) * redfact
    
    def get_bestfit ( self, key ) :
        """ Returns the bestfit value of the stored quantity corresponding 
        to the input key
        
        Parameters
        ----------
        key : str or str-sequence 
            if a string, it should name one of the stored quantities 
            if a list of strings, all the strings in the list will be
            matched.  
        
        Returns
        -------
        : scalar or ndarray
            depending on the input ``key``.
        """
        idmax = self.logl.argmax()
        return func_scalar_or_array(
            var = key,
            function = lambda k : self.__dict__[k][idmax]
        )

    def get_mean ( self, key ) :
        """ Returns the weighted mean value of the stored quantity
        corresponding to the input key
        
        Parameters
        ----------
        key : str or str-sequence 
            if a string, it should name one of the stored quantities 
            if a list of strings, all the strings in the list will be
            matched.  
        
        Returns
        -------
        : scalar or ndarray
            depending on the input ``key``.
        """
        return func_scalar_or_array(
            var = key,
            function = lambda k : numpy.average( self.__dict__[k][self.wnot0],
                                                 weights = self.weights[self.wnot0],
                                                 axis = 0 )
        )

    def get_std ( self, key ) :
        """ Returns the weighted standard deviation value of the stored 
        quantity corresponding to the input key
        
        Parameters
        ----------
        key : str or str-sequence 
            if a string, it should name one of the stored quantities 
            if a list of strings, all the strings in the list will be
            matched.  
        
        Returns
        -------
        : scalar or ndarray
            depending on the input ``key``.
        """
        return func_scalar_or_array(
            var = key,
            function = lambda k : numpy.sqrt(
                numpy.average( ( self.__dict__[k][self.wnot0] - self.get_mean( k ) )**2,
                               weights = self.weights[self.wnot0], axis = 0 )
            )
        )

    def get_quantile ( self, key, quantile = 0.5 ) :
        """ Returns the weighted standard deviation value of the stored 
        quantity corresponding to the input key
        
        Parameters
        ----------
        key : str or str-sequence 
            if a string, it should name one of the stored quantities 
            if a list of strings, all the strings in the list will be
            matched.  
        
        Returns
        -------
        : scalar or ndarray
            depending on the input ``key``.
        """
        return func_scalar_or_array(
            var = key,
            function = lambda k : quantile_weighted( self.__dict__[ k ][self.wnot0],
                                                     quantile,
                                                     weights = self.weights[self.wnot0],
                                                     axis = 0 )
        )

    def get_median ( self, key ) :
        """Returns the median of a stored quantity.
        This is a shortcut for ``Results.get_quantile( key, quantile=0.5 )``
        """
        return self.get_quantile( key, quantile=0.5 )

    def get_credible_interval ( self, key, percent = 0.68, centre = 'bestfit' ) :
        """Returns the credible interval around some position enclosing a user-defined
        probability integral.
        Automatically accounts for upper or lower limits on some parameter.
        If an interval can be defined it returns the lower and upper distances from 
        the centre of the interval.
        If only upper/lower limits can be defined, it returns the position of the given limit.

        Parameters
        ----------
        key : str
            the name of one of the stored quantities.
        percent : float
            (Optional, default = 0.68) probability enclosed by the interval
        centre : str
            (Optional, default = 'bestfit') one among ('bestfit', 'mean', 'median'),
            where to define the centre of the interval.
            Note that, if this is set to 'median', and percent=0.5, this is equivalent
            to call ``Results.get_quantile`` with the argument ``quantile=(0.25, 0.75)``
        
        Returns
        -------
        low, upp : tuple
            If the interval is completely defined by the samples, returns the credible 
            limits around the centre so that the integral of the posterior between
            ``centre-low`` and ``centre+upp`` is equal to ``percent``.
            In the case only upper/lower limits can be defined it returns either
            ``(-numpy.inf, upp)`` for an upper limit, so that the integral between ``-inf``
            and ``upp`` is equal to ``percent``, or ``(low, +numpy.inf)`` for a lower limit,
            so that the integral between ``-inf`` and ``low`` is equal to ``1-percent``.
        """
        
        samples = self.__dict__[key][self.wnot0]

        if centre == 'bestfit' :
            idcentre = self.logl[self.wnot0].argmax()
        elif centre == 'mean' :
            idcentre = find_nearest( samples, self.get_mean( key ) )
        elif centre == 'median' :
            idcentre = find_nearest( samples, self.get_median( key ) )
        else :
            raise RuntimeError(
                "``centre`` argument should be one among ('bestfit', 'mean', 'median')"
            )

        # get limits from the weighted distribution
        low, upp = get_credible_interval(
            samples, idcentre, percent, self.weights[self.wnot0]
        )

        # if upper-limit is found, return upper-limited indefinite interval
        if low is None :
            return -numpy.inf, upp
        # if lower-limit is found, return lower-limited indefinite interval
        if upp is None :
            return low, +numpy.inf
        # return lower and upper credibility distances from centre
        return samples[idcentre]-low, upp-samples[idcentre]
    
    def get_model ( self ) :
        """ Instantiates a model corresponding to the one used for sampling.
        
        Returns
        -------
        : galapy.Galaxy.GXY
        """
        if 'pms_kwargs' in self._mod.keys() :
            return PhotoGXY.load( self._mod )
        else :
            return GXY.load( self._mod )

    def get_handler ( self ) :
        """ Instantiates a handler corresponding to the one used for sampling.
        
        Returns
        -------
        : galapy.GalaxyParameters.ModelParameters
        """
        return ModelParameters.load( self._han )
    
    def get_observation ( self ) :
        """ Instantiates a dataset corresponding to the one used for sampling.
        
        Returns
        -------
        : galapy.sampling.Observation.Observation
        """
        if self._obs is not None :
            return Observation.load( self._obs )
        else :
            warnings.warn( "The current instance has no Observation stored, passing None" )
            return None
    
    def get_noise ( self ) :
        """ Instantiates a noise model corresponding to the one used for sampling.
        
        Returns
        -------
        : galapy.sampling.Noise.Noise
        """
        if hasattr(self, '_noise') and self._noise is not None :
            return CalibrationError.load( self._noise )
        else :
            warnings.warn( "The current instance has no Noise stored, passing None" )
            return None
    
    def get_sampling_params ( self ) :
        """ Instantiates a handler corresponding to the one used for sampling.
        
        Returns
        -------
        : galapy.GalaxyParameters.GXYParameters
        """
        return self.get_handler()

#############################################################################################



"""GalaPy models the Spectral Energy Distribution (SED) of galaxies from the X-ray to the Radio band.
 
It is an API as well as the emission itself, models the evolution of 
galaxy components and dust attenuation/reradiation. 
GalaPy provides functions, classes and terminal commands for Bayesian inference of 
galaxy properties from panchromatic photometric data, as well as for the analysis 
and simulation of galaxies spectral properties.
"""

from galapy.configuration import rcParams

from galapy import StarFormationHistory
from galapy import CompositeStellarPopulation
from galapy import InterStellarMedium
from galapy import NebularFreeFree
from galapy import Synchrotron
from galapy import XRayBinaries
from galapy import ActiveGalacticNucleus
from galapy import Cosmology
from galapy import Galaxy
from galapy import InterGalacticMedium
from galapy import PhotometricSystem
from galapy import Noise
from galapy import Handlers
from galapy import sampling, internal, io, analysis

__all__ = [
    'StarFormationHistory', 'CompositeStellarPopulation', 'InterStellarMedium',
    'NebularFreeFree', 'Synchrotron', 'XRayBinaries', 'ActiveGalacticNucleus',
    'Cosmology', 'Galaxy', 'InterGalacticMedium', 'PhotometricSystem',
    'Noise', 'Handlers',
    'sampling', 'internal', 'io', 'analysis',
]
__version__ = 'v0.5.0'



""" GalaPy module handling Star Formation History (SFH) models.
"""

# External imports
import numpy

# Internal imports
from .SFH_core import CSFH
#from .internal.utils import gxyComp

sfh_tunables = {
    # In-Situ SF model
    'insitu' : ( 'psi_max', 'tau_star' ),
    
    # Constant SF model
    'constant' : ( 'psi', 'Mdust', 'Zgxy' ),
    
    # Delayed-Exponential model
    'delayedexp' : ('psi_norm', 'k_shape', 'tau_star', 'Mdust', 'Zgxy' ),
    
    # Log-Normal model
    'lognormal' : ('psi_norm', 'sigma_star', 'tau_star', 'Mdust', 'Zgxy' ),
    
    # Interpolated model
    'interpolated' : ('Mdust', 'Zgxy' )
}
""" Dictionary of tunable parameters
"""
    
def sfh_build_params ( tau_quench = 2.e+10, model = 'insitu', **kwargs ) :
    """ Builds the parameter dictionary for a given SFH model.

    Parameters
    ----------
    tau_quench : float
       eventual time of quenching in units of year, 
       defaults to the arbitrary large value of $2 \cdot 10^9$ years
    model : string
       SFH model
    **kwargs : dictionary, optional
       The free-parameters of the SFH model chosen (with argument ``model``)
       The parameterization depends on the chosen model.
    
    Keyword Arguments
    -----------------
    psi_max : float
      ``model = 'insitu'``, optional, default = 100
    tau_star : float
      ``model = ['insitu','delayedexp', 'lognormal']``, optional, 
      default = 3.e+8 if ``insitu`` or ``lognormal``, = 1.e+8 if ``delayedexp``
    psi : float
      ``model = 'constant'``, optional, default = 1.0
    psi_norm : float
      ``model = ['delayedexp', 'lognormal']``, optional, 
      default = 1.0 if ``delayedexp``, = 100 if ``lognormal``
    k_shape : float
      ``model = 'delayedexp'``, optional, default = 0.2
    sigma_star : float
      ``model = 'lognormal'``, optional, default = 2.0
    Mdust : float
      ``model = ['constant','delayedexp','lognormal','interpolated']``, optional, default = 1.e+8
    Zgxy : float
      ``model = ['constant','delayedexp','lognormal','interpolated']``, optional, default = 0.01

    Returns
    -------
    : dict
       Dictionary containing the parameterization of the chosen SFH model.
       All the parameters that have not been passed to the function are defined
       with their default value.
    """

    _models = {
        # In-Situ SF model
        'insitu' : {
            'psi_max'  : 100.,
            'tau_star' : 3.e+8,
        },
        # Constant SF model
        'constant' : {
            'psi'   : 1.,
            'Mdust' : 1.e+8,
            'Zgxy'  : 0.01,
        },
        # Delayed-Exponential model
        'delayedexp' : {
            'psi_norm' : 1.,
            'k_shape'  : 0.2,
            'tau_star' : 1.e+8,
            'Mdust'    : 1.e+8,
            'Zgxy'     : 0.01,
        },
        # Log-Normal model
        'lognormal' : {
            'psi_norm'   : 100.,
            'sigma_star' : 2.,
            'tau_star'   : 3.e+8,
            'Mdust'      : 1.e+8,
            'Zgxy'       : 0.01,
        },
        # Interpolated model
        'interpolated' : {
            'Mdust'      : 1.e+8,
            'Zgxy'       : 0.01,
        }
    }
    out = { 'tau_quench' : tau_quench, 'model' : model }
    out.update( _models[ model ] )
    for k in set(out.keys()).intersection(kwargs.keys()) :
        out[k] = kwargs[k]

    return out
        

class SFH () :
    r""" Class wrapping the C-core implementation of the Star Formation History type.   
 
    The possible models to choose are:

    #. ``'insitu'``
       
       .. math ::
          
          \psi(\tau) = \psi_\text{max} \left[ \exp(-x) - \exp(-s\gamma x) \right]
       
       with :math:`x = \tau / (s\tau_\star)`, :math:`s=3`, 
       :math:`\gamma = 1 - \mathcal{R} + 3 \psi_\text{max}^{-0.3}`, 
       :math:`\mathcal{R} = 0.45` the instantaneous recycling factor, 
       :math:`\tau_\star =`:code:`tau_star` and :math:`\psi_\text{max} =`:code:`psi_max`  
    
    #. ``'constant'``
       
       .. math ::
          
          \psi(\tau) = \psi_0
       
       with :math:`\psi_0 =`:code:`psi`
    
    #. ``'delayedexp'``
       
       .. math ::
          
          \psi(\tau) = \psi_\text{norm} \tau^{\kappa}\, \exp{(-\tau/\tau_\star)}
       
       with :math:`\psi_\text{norm} =`:code:`psi_norm`, 
       :math:`\kappa =`:code:`k_shape` and :math:`\tau_\star =`:code:`tau_star`
    
    #. ``'lognormal'``
       
       .. math ::
          
          \psi(\tau) = \psi_\text{norm} \dfrac{1}{\tau}\, \dfrac{1}{\sqrt{2\pi\sigma_\star^2}}\, 
          \exp\left[-\dfrac{\ln^2(\tau/\tau_\star)}{2\,\sigma_\star^2}\right]
       
       with :math:`\psi_\text{norm} =`:code:`psi_norm`, :math:`\sigma_\star =`:code:`sigma_star` 
       and :math:`\tau_\star =`:code:`tau_star`
    
    #. ``'interpolated'``: from a pre-computed grid of SFRs computed at given times.
       If this model is chosen, two positional arguments are mandatory, 
       i.e. the two grids ``args = (tau,sfr)`` with ``len(tau)==len(sfr)``, see below.

    Parameters
    ----------
    *args :
      positional arguments are optional, if passed they should be the two-iterables 
      corresponding to gridded values of time and SFR to interpolate over: ``args = (tau, sfr)``
      with ``len(tau)==len(sfr)``. Note that in this case the 'model' argument will be over-written
      and fixed to 'interpolated'
    tau_quench : float
      Eventual abrupt quenching time for star formation. 
      Should be expressed in years. It refers to the age of the galaxy,
      i.e. it has to be intended as the time passed from the formation of the galaxy.

    model : string
      One among ( 'insitu', 'constant', 'delayedexp', 'lognormal', 'interpolated', 'burst' ).
      Default is 'insitu'.

    **kwargs : dictionary, optional
       The free-parameters of the SFH model chosen (with argument ``model``)
       The parameterization depends on the chosen model.
    
    Keyword Arguments
    -----------------
    psi_max : float
      ``model = 'insitu'``, optional, default = 100
    tau_star : float
      ``model = ['insitu','delayedexp', 'lognormal']``, optional, 
      default = 3.e+8 if ``insitu`` or ``lognormal``, = 1.e+8 if ``delayedexp``
    psi : float
      ``model = 'constant'``, optional, default = 1.0
    psi_norm : float
      ``model = ['delayedexp', 'lognormal']``, optional, 
      default = 1.0 if ``delayedexp``, = 100 if ``lognormal``
    k_shape : float
      ``model = 'delayedexp'``, optional, default = 0.2
    sigma_star : float
      ``model = 'lognormal'``, optional, default = 2.0
    Mdust : float
      ``model = ['constant','delayedexp','lognormal','interpolated']``, optional, default = 1.e+8
    Zgxy : float
      ``model = ['constant','delayedexp','lognormal','interpolated']``, optional, default = 0.01

    Note
    ----
    Not for SED fitting, use the galaxy class
    """

    def __init__ ( self, *args, tau_quench = 2.e+20, model = 'insitu', **kwargs ) :

        if len(args) == 2 :
            try :
                tau = numpy.ascontiguousarray(args[0])
                sfr = numpy.ascontiguousarray(args[1])
                model = 'interpolated'
            except :
                raise;
        if 'tau' in kwargs and 'sfr' in kwargs :
            try :
                tau = numpy.ascontiguousarray(kwargs['tau'])
                sfr = numpy.ascontiguousarray(kwargs['sfr'])
                model = 'interpolated'
            except :
                raise;            
            
        self.core = CSFH( tau_quench, model )
        self.params = sfh_build_params( tau_quench, model, **kwargs )
        self.tunable = set( self.params.keys() )
        self.tunable.remove('model')
        if model == 'interpolated' :
            self.core.set_interpolator(tau, sfr)
        self.set_parameters()

    def __call__ ( self, tau ) :
        """ Method for returning the Star Formation Rate (SFR) at given age.

        Parameters
        ----------
        tau : float or array-like
          The time in the galaxy SFH in which to compute the SFR

        Returns
        -------
        SFR : float or array-like
          SFR at given time from galaxy formation.
        """
        tau = numpy.asarray( tau )
        scalar = False
        if tau.ndim == 0 :
            tau = tau[None] # makes 'tau' 1D
            scalar = True
        ret = self.core( tau )
        if scalar :
            return ret.item()
        return ret

    def set_parameters ( self, tau_quench = None, **kwargs ) :
        r""" Function for setting the parameters of the model.

        Parameters
        ----------
        tau_quench : float
          Eventual time of abrupt quenching event, stopping 
          star formation forever.
        **kwargs : dictionary, optional
          The free-parameters of the SFH model chosen (with argument ``model``)
          The parameterization depends on the chosen model.
    
        Keyword Arguments
        -----------------
        psi_max : float
          ``model = 'insitu'``, optional, default = 100
        tau_star : float
          ``model = ['insitu','delayedexp', 'lognormal']``, optional, 
          default = 3.e+8 if ``insitu`` or ``lognormal``, = 1.e+8 if ``delayedexp``
        psi : float
          ``model = 'constant'``, optional, default = 1.0
        psi_norm : float
          ``model = ['delayedexp', 'lognormal']``, optional, 
          default = 1.0 if ``delayedexp``, = 100 if ``lognormal``
        k_shape : float
          ``model = 'delayedexp'``, optional, default = 0.2
        sigma_star : float
          ``model = 'lognormal'``, optional, default = 2.0
        Mdust : float
          ``model = ['constant','delayedexp','lognormal','interpolated']``, optional, default = 1.e+8
        Zgxy : float
          ``model = ['constant','delayedexp','lognormal','interpolated']``, optional, default = 0.01
        
        Returns
        -------
        : None
        """
        if tau_quench is not None :
            self.core.set_tau_quench( tau_quench )
            self.params['tau_quench'] = tau_quench
        self.params.update( kwargs )
        self.core.set_params( numpy.asarray( [
            self.params[k]
            for k in sfh_tunables[self.params['model']]
        ], dtype=float) )
        return;

    def Mstar ( self, tau, npoints = 100 ) :
        """ Computes the stellar mass at a given age of the galaxy.
        
        It approximates the integral:

        .. math::
        
          M_\\ast(\\tau') = \\int_0^{\\tau'}\\text{d}\\tau 
                            \\bigl[1 - \\mathcal{R}_\\text{IMF}(\\tau)\\bigr]\\psi(\\tau)

        Parameters
        ----------
        tau : float or array-like of floats
          galaxy age in years.
        npoints : int
          thinness for approximated integral computation (default is 100)

        Returns
        -------
        : float or array-like of floats
          the stellar mass of the galaxy at time :math:`\\tau`        
        """

        tau = numpy.asarray( tau )
        scalar_input = False
        if tau.ndim == 0 :
            tau = tau[None] # makes 'tau' 1D
            scalar_input = True
        ret = numpy.asarray( [ self.core.Mstar( _t, npoints )
                               for _t in tau ],
                             dtype=numpy.float64 )
        
        if scalar_input :
            return ret.item()
        return ret

    def Mdust ( self, tau ) :
        """ Returns the dust mass at a given age of the galaxy.
  
        For empirical models of star formation (i.e. :code:`const`, 
        :code:`delayedexp`, :code:`lognorm`) this is a free parameter.

        For the :code:`insitu` model, the dust mass is given by 

        .. math::
           
           M_\\text{dust} = M_\\text{gas}(\\tau)D(\\tau)

        where 

        .. math::
           
           M_\\text{gas}=\\psi(\\tau)\\tau_\\ast
 
        and where :math:`D(\\tau)` is the gas mass ratio 
        (for an analytic expression of this quantity 
        see Pantoni et al. 2019 and Lapi et al. 2020).

        Parameters
        ----------
        tau : float or array-like of floats
          age of the galaxy in years.

        Returns
        -------
        : float or array-like of floats
          Dust content in solar masses (:math:`M_\\odot`) at give time :math:`\\tau`.
        """
        
        tau = numpy.asarray( tau )
        scalar_input = False
        if tau.ndim == 0 :
            tau = tau[None] # makes 'tau' 1D
            scalar_input = True
        ret = numpy.asarray( [ self.core.Mdust( _t )
                               for _t in tau ],
                             dtype=numpy.float64 )
        
        if scalar_input :
            return ret.item()
        return ret

    def Mgas ( self, tau ) :
        """ Returns the gas mass at a given age of the galaxy.

        For empirical models of star formation (i.e. :code:`const`, 
        :code:`delayedexp`, :code:`lognorm`) 
        this is given by 

        .. math::
           
           M_\\text{gas} = M_\\text{dust}/D
        
        where :math:`D \\sim 0.01 (Z_\\text{gas}/Z_\\odot)^{-0.85}` is the 
        dust-to-gas mass ratio, derived from observations.

        For the :code:`insitu` model, the gas mass is given by

        .. math::
           
           M_\\text{gas}=\\psi(\\tau)\\tau_\\ast

        Parameters
        ----------
        tau : float or array-like of floats
          age of the galaxy in years.

        Returns
        -------
        : float or array-like of floats 
          Gas content in solar masses (:math:`M_\\odot`) at give time :math:`\\tau`.
        """
        
        tau = numpy.asarray( tau )
        scalar_input = False
        if tau.ndim == 0 :
            tau = tau[None] # makes 'tau' 1D
            scalar_input = True
        ret = numpy.asarray( [ self.core.Mgas( _t )
                               for _t in tau ],
                             dtype=numpy.float64 )
        
        if scalar_input :
            return ret.item()
        return ret

    def Zgas ( self, tau ) :
        """ Returns the gas metallicity at a given age of the galaxy.
        
        For empirical models of star formation (i.e. :code:`const`, 
        :code:`delayedexp`, :code:`lognorm`) this is a free parameter with 
        :math:`Z_\\text{gas} = Z_\\ast`.

        For the :code:`insitu` model, it is instead given by

        .. math::

          Z_\\text{gas}=\\dfrac{s y_Z}{s\\gamma-1}
                        \\biggl[
                           1 - \\dfrac{(s\\gamma-1)x}{e^{(s\\gamma-1)x}-1}
                        \\biggr]
    
        where :math:`x\\equiv\\tau/s\\tau_\\ast` and :math:`y_Z\\approx0.04` 
        is the metal production yield (including recycling) for a Chabrier IMF.

        Parameters
        ----------
        tau : float or array-like of floats
          age of the galaxy in years.
    
        Returns
        -------
        : float or array-like of floats
          Gas absolute metallicity at give time :math:`\\tau`.
        """
    
        tau = numpy.asarray( tau )
        scalar_input = False
        if tau.ndim == 0 :
            tau = tau[None] # makes 'tau' 1D
            scalar_input = True
        ret = numpy.asarray( [ self.core.Zgas( _t )
                               for _t in tau ],
                             dtype=numpy.float64 )
        
        if scalar_input :
            return ret.item()
        return ret

    def Zstar ( self, tau ) :
        """Returns the stellar metallicity at a given age of the galaxy.

        For empirical models of star formation (i.e. :code:`const`, 
        :code:`delayedexp`, :code:`lognorm`) this is a free parameter with 
        :math:`Z_\\ast = Z_\\text{gas}`.

        For the :code:`insitu` model, it is instead given by

        .. math::

          Z_\\ast=\\dfrac{y_Z}{\\gamma-1}
                  \\biggl[1 - \\dfrac{s\\gamma}{(s\\gamma-1}
                  \\dfrac{e^{-x}-e^{-s\\gamma x}[1 + (s\\gamma -1)x]}
                  {s\\gamma -1 + e^{-s\\gamma x}- s\\gamma e^{-x}}\\biggr]

        where :math:`x\\equiv\\tau/s\\tau_\\ast` and :math:`y_Z\\approx0.04` 
        is the metal production yield (including recycling) for a Chabrier IMF.

        Parameters
        ----------
        tau : float or array-like of floats
          age of the galaxy in years.

        Returns
        -------
        : float or array-like of floats
          Stellar absolute metallicity at give time :math:`\\tau`.
        """
        
        tau = numpy.asarray( tau )
        scalar_input = False
        if tau.ndim == 0 :
            tau = tau[None] # makes 'tau' 1D
            scalar_input = True
        ret = numpy.asarray( [ self.core.Zstar( _t )
                               for _t in tau ],
                             dtype=numpy.float64 )
        
        if scalar_input :
            return ret.item()
        return ret
    



""" Noise module
"""

# External imports
import numpy
from abc import abstractmethod
from collections.abc import MutableMapping as MM

# Internal imports
from galapy.internal.abc import Model

########################################################################################

noise_params_defaults = {

    # Calibration error
    'f_cal' : ( 'Fractional error done in calibrating data',
                [-10., 1], True, 'f_\\mathrm{cal}' )

}

########################################################################################
# Base abstract class

class Noise ( Model ) :
    
    def __init__ ( self ) :
        super().__init__()

    @abstractmethod
    def dump ( self ) :
        pass
        
    @abstractmethod
    def apply ( self ) :
        pass

########################################################################################
# Calibration error

class CalibrationError ( Noise ) :
    
    def __init__ ( self, f_cal = 0.0 ) :
        """
        """
        super().__init__()
        self.f_cal = f_cal
        self.params['f_cal'] = self.f_cal

    def dump ( self ) :
        return self.params

    @classmethod
    def load ( cls, dictionary ) :
        return cls( **dictionary )
        
    def set_parameters ( self, f_cal = None ) :
        """
        """
        if f_cal is not None :
            self.f_cal = f_cal
            self.params['f_cal'] = self.f_cal
        return;
        
    def apply ( self, error, model ) :
        """
        """
        error = numpy.asarray( error )
        model = numpy.asarray( model )
        if len(error) != len(model) :
            raise AttributeError( 'Attributes error and model should have same length' )
        
        return numpy.sqrt( error**2 + ( model * self.f_cal )**2 )

########################################################################################




""" The XRayBinaries module implements the contribution to the X-band emission from stellar binary systems.
"""

# External imports
from abc import ABC, abstractmethod
import numpy

# Internal imports
from galapy.internal.utils import trap_int, powerlaw_exp_cutoff, poly_N
from galapy.internal.constants import Ang_to_keV, Lsun, sunL
from galapy.internal.interp import lin_interp

xrb_tunables = {
    # High Mass Stars
    'hm' : [ 'psi', 'Zstar' ],
    
    # Low Mass Stars
    'lm' : [ 'Mstar', 'age' ],
}
""" Dictionary with the tunable parameters of the X-ray contribution from binary star systems

#. :code:`hm` : **High Mass Star**
#. :code:`lm` : **Low Mass Stars**

"""

def xrb_build_params ( startype, **kwargs ) :
    """ Builds the parameters dictionary for given star type.
    """
    if startype not in xrb_tunables.keys() :
        av_types = ''
        for k in xrb_tunables.keys() :
            av_types += f'"{k}" '
        raise ValueError( f'Star type "{startype}" chosen not available. '
                          f'Available phases are: {av_types}')
    
    if startype == 'hm' :
        out = {
            'psi'   : 1.,
            'Zstar' : 0.02,
        }
        
    if startype == 'lm' :
        out = {
            'Mstar' : 1.e+10,
            'age'   : 1.e+8,
        }

    for k in set( out.keys() ).intersection(kwargs.keys()) :
        out[k] = kwargs[k]
    return out

class _XRBase ( ABC ) :

    @property
    @abstractmethod
    def coeff ( self ) :
        pass

    def __init__ ( self, startype, gamma, Ecut,
                   lmin = 1., lmax = 1.e+10, **kwargs ) :
        self.startype = startype
        self.params   = xrb_build_params( self.startype, **kwargs )
        self.lmin     = lmin
        self.lmax     = lmax
        self.NormStarType = None

        ###########################################################
        # Compute X-template

        # generate wavelength grid
        ll = numpy.logspace( numpy.log10( self.lmin ),
                             numpy.log10( self.lmax ),
                             256 )
        
        # convert wavelength to energy
        El = Ang_to_keV( ll )
        
        # find interval for hard-X normalization
        wE = ( 2. <= El ) & ( El <= 10. )

        # compute emission law normalization
        Lnorm = -1. / trap_int( El[wE], powerlaw_exp_cutoff( El[wE], gamma=gamma, Ecut=Ecut ) )

        # Compute the normalized power-law:
        # - high energy cut-off at 300 keV
        # - low energy cut-off at 50 Angstrom (~0.25 keV)
        ret = numpy.zeros_like( ll )
        wL = ( ll <= 5.e+1 ) 
        ret[wL] = powerlaw_exp_cutoff( El[wL], gamma=gamma, Ecut=Ecut ) * Lnorm

        # store normalized interpolator object for X-ray emission
        self.f_norm_X = lin_interp( numpy.ascontiguousarray( ll ),
                                    numpy.ascontiguousarray( ret ) )

    @abstractmethod
    def set_parameters ( self ) :
        pass
    
    def _set_norm ( self, xx, coeff, fact ) :
        logNorm = poly_N( xx, coeff )
        self.NormStarType = fact * 10.**logNorm * sunL
        return;

    def emission ( self, ll ) :
        return self.NormStarType * self.f_norm_X( ll )

    
class HMXRB ( _XRBase ) :

    coeff = [ 1968.33, -1883.80, 569.44, -62.12, 40.28 ]

    def __init__ ( self, lmin, lmax, **kwargs ) :
        super().__init__( startype = 'hm',
                          gamma    = 2.0,
                          Ecut     = 100.,
                          lmin = lmin, lmax = lmax,
                          **kwargs )
        self.set_parameters()

    def set_parameters ( self, **kwargs ) :

        self.params.update( kwargs )
        super()._set_norm( self.params['Zstar'],
                           type( self ).coeff,
                           self.params['psi'] )
        return;

    
class LMXRB ( _XRBase ) :

    coeff = [ 0.136, 0.425, -0.423, -1.503, 40.276 ]

    def __init__ ( self, lmin, lmax, **kwargs ) :
        super().__init__( startype = 'lm',
                          gamma    = 1.6,
                          Ecut     = 100.,
                          lmin = lmin, lmax = lmax,
                          **kwargs )
        self.set_parameters()

    def set_parameters ( self, **kwargs ) :

        self.params.update( kwargs )
        super()._set_norm( numpy.log10( self.params['age'] * 1.e-9 ),
                           type( self ).coeff,
                           self.params['Mstar'] * 1.e-10 )
        return;

class XRB () :

    def __init__ ( self, lmin, lmax, **kwargs ) :

        self.params = {}
        self.hm = HMXRB(
            lmin, lmax,
            **{ k : v
                for k, v in kwargs.items()
                if k in xrb_tunables[ 'hm' ] }
        )
        self.params.update( self.hm.params )
        
        self.lm = LMXRB(
            lmin, lmax,
            **{ k : v
                for k, v in kwargs.items()
                if k in xrb_tunables[ 'lm' ] }
        )
        self.params.update( self.lm.params )

    def set_parameters ( self, **kwargs ) :
        
        self.params.update( kwargs )
        self.hm.set_parameters( **{ k : v
                                    for k, v in kwargs.items()
                                    if k in xrb_tunables[ 'hm' ] } )
        self.lm.set_parameters( **{ k : v
                                    for k, v in kwargs.items()
                                    if k in xrb_tunables[ 'lm' ] } )
        return;

    def emission ( self, ll ) :
        return (
            self.hm.emission( ll ) +
            self.lm.emission( ll )
        )



import numpy
from collections.abc import MutableMapping, Sequence
from galapy.PhotometricSystem import PMS

class Observation () :
    
    def __init__ ( self, bands, fluxes, errors, uplims, pms ) :
        """ Class for managing photometric observations of galaxies.
        It builds the photometric system from arguments and stores the 
        observational data in ordered numpy arrays. The order is set
        by the photometric system chosen. By construction, it follows
        the pivot wavelengths of the transmission filters provided.
        
        Parameters
        ----------
        bands : array-like
          sequence of strings containing the name of the bandpass 
          filters associated to the obseved dataset

        fluxes : array-like
          sequence of fluxes (or upper limits) measured
          in each of the bandpass filters listed in argument `bands`

        errors : array-like
          sequence of errors associated to the fluxes (or upper limits)
          listed in argument `fluxes`

        uplims : array-like
          sequence of booleans identifying whether the values listed 
          in argument `fluxes` should be considered non-detection (`True`)
          or a detection (`False`)

        pms : galapy.PhotometricSystem.PMS
          instance of type ``galapy.PhotometricSystem.PMS`` with 
          the collection of bandpass transmissions producing the 
          integrated fluxes for the observation (same as listed in
          argument ``bands``)
        """

        if not isinstance( pms, PMS ) :
            try :
                self.pms = PMS( *bands )
            except Exception as err :
                print(
                    'Not possible to deduce the photometric system '
                    'from the list of bands provided. Exited with exception: '
                    f'{err:s}' )
                raise
            else :
                raise RuntimeError(
                    'pms argument must be an instance of type ``galapy.PhotometricSystem.PMS``'
                )
        self.pms = pms
            
        if any( [ len( lst ) != len( self.pms ) 
                  for lst in [ bands, fluxes, errors, uplims ] ] ) :
            raise ValueError( 'All arguments should have the same length.' )
        
        # create temporary data dictionary to iterate by-key in the order provided by the pms object
        data = { k : v for k, v in zip( bands, zip( fluxes, errors, uplims ) ) }
        
        # store internally the dataset with the same order
        # of the bands defined in the photometric system
        self.fluxes = numpy.asarray( [ data[k][0] for k in self.pms.keys ] )
        self.errors = numpy.asarray( [ data[k][1] for k in self.pms.keys ] )
        self.uplims = numpy.asarray( [ data[k][2] for k in self.pms.keys ] )

    def dump ( self ) :
        return dict(
            bands  = '|'.join(self.pms.keys),
            fluxes = self.fluxes,
            errors = self.errors,
            uplims = self.uplims,
            pms_kwargs = self.pms.dump(),
        )

    @classmethod
    def load ( cls, dictionary ) :
        # ensure deep-copy
        dictionary = dict( dictionary )
        return cls(  
            dictionary.pop('bands').split('|'),
            pms = PMS.load( dictionary.pop( 'pms_kwargs' ) ),
            **dictionary
        )



# External imports
# from abc import ABC, abstractmethod
from collections.abc import MutableMapping as MM
import numpy

###################################################################################

def trap_int ( xx, yy ) :
    r""" Trapezoid integration

    Parameters
    ----------
    xx : array-like
       x-domain grid
    yy : array-like
       y-domain grid
    
    Returns
    -------
    : float 
       the integral along the whole x-domain
    """
    xx = numpy.asarray(xx)
    yy = numpy.asarray(yy)
    return numpy.sum( 0.5 * ( yy[:-1] + yy[1:] ) * ( xx[1:] - xx[:-1] ), axis = 0 )

###################################################################################

class FlagVal () :
    def __init__ ( self, value, flag ) :
        self.value = value
        self.flag  = flag
    def __repr__ ( self ) :
        return f'{type(self).__name__}({self.value},"{self.flag}")'
    def __str__ ( self ) :
        return f'{self.flag}:{self.value}'

###################################################################################

def find_nearest ( array, value ) :
    r""" Function finding the indexes of array closer to the
    values in value.
    
    Parameters
    ----------
    array : array-like 
    
    value : array-like or scalar
    
    Return
    ------
    : numpy-array
      list of indexes of elements in array closer to 
      values in value.
    
    Warning
    -------
    Uniqueness of elements in the returned array is not guaranteed.
    """
    import numpy
    
    array = numpy.asarray( array )
    value = numpy.asarray( value )
    scalar_input = False
    if value.ndim == 0 :
        value = value[None] # makes 'value' 1D
        scalar_input = True
        
    idx = [ ( numpy.abs( array - _v ) ).argmin() for _v in value ] 

    if scalar_input :
        return numpy.squeeze( idx )
    return numpy.asarray( idx )

###################################################################################

def powerlaw_exp_cutoff ( El, gamma, Ecut ) :
    r""" Returns a powerlaw with exponential cut-off of the form

    .. math::
      
      y = E_\lambda^{-\gamma+3} e^{E_\lambda/E_\text{cut}}

    Parameters
    ----------
    El : scalar or array-like
      the x-values
    gamma : scalar
      the spectral index
    Ecut : scalar
      the cut-off x-scale
    
    Returns
    -------
    : scalar or array-like
    """
    return El**(-gamma+3) * numpy.exp(-El/Ecut) 

###################################################################################

def poly_N ( xx, coeff ) :
    r""" Method for computing N-order polynomia.
    The order of the polynomium is set by the length of the `coeff` list.
    
    The method computes

    .. math::
    
       y = \sum_0^N c_i * x^{N-i}
    
    Parameters
    ----------
    xx : scalar or array-like
      the x-values 
    coeff : list
      the values of the coefficients of the polynomium. The list must be
      ordered, with the first element corresponding to the coefficient multiplying
      the highest power of the x-variable and the last coefficient being the 
      scalar coefficient
    
    Returns
    -------
    : scalar or array-like
      the y-values
    """
    yy = 0.
    for _c in coeff :
        yy = yy * xx + _c
    return yy

###################################################################################

def unwrap_keys ( d ) :
    r""" Generator yielding the keys of a nested dictionary

    Parameters
    ----------
    d : dictionary
      A nested dictionary (whathever object inheriting from 'dict' 
      or collections.abc.MutableMapping or similar)
    
    Yields
    ------
    : list of keys    

    Examples
    --------
    >>> d = { 'a' : 1, 'b' : { 'c' : 2, 'd' : 3 } }
    >>> for i in unwrap_keys( d ) : print( i )
    ['a']
    ['b', 'c']
    ['b', 'd']
    """
    from collections.abc import MutableMapping
    for k, v in d.items() :
        if isinstance( v, MutableMapping ) :
            for t in unwrap_keys( v ) :
                yield [ k ] + t
        else :
            yield [ k ]

def unwrap_values ( d ) :
    r""" Generator yielding the values of a nested dictionary

    Parameters
    ----------
    d : dictionary
      A nested dictionary (whathever object inheriting from 'dict' 
      or collections.abc.MutableMapping or similar)
    
    Yields
    ------
    : values

    Examples
    --------
    >>> d = { 'a' : 1, 'b' : { 'c' : 2, 'd' : 3 } }
    >>> for i in unwrap_values( d ) : print( i )
    1
    2
    3
    """
    from collections.abc import MutableMapping
    for k, v in d.items() :
        if isinstance( v, MutableMapping ) :
            for t in unwrap_values( v ) :
                yield t
        else :
            yield v

def unwrap_items ( d ) :
    r""" Generator yielding the items of a nested dictionary

    Parameters
    ----------
    d : dictionary
      A nested dictionary (whathever object inheriting from 'dict' 
      or collections.abc.MutableMapping or similar)
    
    Yields
    ------
    : items
      a 2d tuple in which the first element is the list of nested keys
      and the second is the associated value

    Examples
    --------
    >>> d = { 'a' : 1, 'b' : { 'c' : 2, 'd' : 3 } }
    >>> for i in unwrap_items( d ) : print( i )
    (['a'], 1)
    (['b', 'c'], 2)
    (['b', 'd'], 3)
    """
    from collections.abc import MutableMapping
    for k, v in d.items() :
        if isinstance( v, MutableMapping ) :
            for tk, tv in unwrap_items( v ) :
                yield [ k ] + tk, tv
        else :
            yield [ k ], v

###################################################################################

def set_nested ( d, kl, v ) :
    r""" Recursive function for setting values in a nested dictionary.
    
    Parameters
    ----------
    d : dictionary
        the i/o dictionary
    kl : list
        a list of keywords to traverse the nested dictionary
        the first element is the highest position in the hierarchy
        of the nested dictionary
    v : whatever
        the value to set at corresponding nested sequence of keywords
    
    Returns
    -------
    : None
    
    Examples
    --------
    >>> d = {}
    >>> kl1 = ['first', 'second', 'third']
    >>> kl2 = ['a', 'b', 'c']
    >>> set_nested(d,kl1, 3.)
    >>> set_nested(d,kl2, True)
    >>> d
    {'first': {'second': {'third': 3.0}}, 'a': {'b': True}}
    
    If we want to add a new key-value pair to an already existing level:
    
    >>> kl3 = ['first', 'second', 'fourth']
    >>> set_nested(d, kl3, 4.)
    >>> d
    {'first': {'second': {'third': 3.0, 'fourth': 4.0}},'a': {'b': True}}
    """
    k = kl.pop(0)
    if len(kl) == 0 :
        d[k] = v
        return;
    try :
        set_nested( d[k], kl, v )
    except KeyError :
        d[k] = {}
        set_nested( d[k], kl, v )
    return;        

###################################################################################

def get_nested ( d, kl ) :
    if len( kl ) > 1 and isinstance( d, MM ) :
        return get_nested( d[ kl.pop( 0 ) ], kl )
    return d[ kl.pop( 0 ) ]

###################################################################################

def func_scalar_or_array ( var, function, *args, **kwargs ) :
    """ A function applying some function to a scalar or an ndarray.
    
    Parameters
    ----------
    var : scalar or ndarray
        The argument of the function
    function : callable
        A function taking a scalar as input
    args : sequence
        A list or tuple of additional arguments to be passed to function
    kwargs : dict
        A dictionary of additional keyword arguments to be passed to function
    
    Returns
    -------
    : scalar or ndarray
        Depends on the shape of ``var``: (i) if ``var`` is a scalar it is the result 
        of ``function(var)``; (ii) if ``var`` is an ndarray it is the result of 
        ``array([function(v) for v in var])``
    """
    var = numpy.asarray(var)
    scalar = False
    if var.ndim == 0 :
        var = var[None]
        scalar = True
    ret = numpy.array([function(v,*args,**kwargs) for v in var])
    if scalar :
        return ret[0]
    return ret

###################################################################################

def quantile_weighted ( values, quantiles, weights = None, 
                        values_sorted = False, old_style = False, 
                        axis=-1 ) :
    """ Very close to numpy.percentile, but supports weights.
    (partially stolen from stackoverflow: 
    https://stackoverflow.com/questions/21844024/weighted-percentile-using-numpy)

    Parameters
    ----------
    values : numpy.array 
        with data
    quantiles : array-like 
        quantiles needed (NOTE: quantiles should be in [0, 1]!)
    sample_weight : array-like 
        same length as ``values``
    values_sorted : bool
        if True, then will avoid sorting of initial array
    old_style : bool
        if True, will correct output to be consistent
        with numpy.percentile.
    axis : int
    
    Returns
    -------
    : numpy.array 
        computed quantiles.
    """
    values = numpy.array(values)
    quantiles = numpy.array(quantiles)
    if weights is None:
        weights = numpy.ones(len(values))
        
    weights = numpy.array(weights)
    if numpy.any(quantiles < 0) or numpy.any(quantiles > 1) :
        raise ValueError( 'Quantiles should be in [0, 1]' )        

    def get_weighted_1D( val, wgh ) :
        if not values_sorted :
            sorter = numpy.argsort(val)
            val = val[sorter]
            wgh = weights[sorter]
            
        weighted_quantiles = numpy.cumsum(wgh) - 0.5 * wgh
        if old_style : 
            # To be convenient with numpy.percentile
            weighted_quantiles -= weighted_quantiles[0]
            weighted_quantiles /= weighted_quantiles[-1]
        else :
            weighted_quantiles /= numpy.sum(wgh)
        return numpy.interp(quantiles, weighted_quantiles, val)
    
    if values.ndim > 1 :
        axes = numpy.delete(numpy.arange(values.ndim, dtype=numpy.int64), axis)
        res = numpy.empty( ( *[values.shape[a] for a in axes], *quantiles.shape ) )
        values = numpy.transpose(values, axes = numpy.append( axes, axis ))
        for i, val in enumerate(values) :
            res[i] = get_weighted_1D( val, weights )

        return res.T
    
    return get_weighted_1D( values, weights )

###################################################################################

def get_credible_interval(samples, idcentre, percent = 0.68, weights = None) :
    """Find lower and upper limits in sample array containing 
    given percent value around position.
    
    It splits the samples in 2 half-intervals:
    [lower, samples[idcentre]) and [samples[idcentre], upper) that are set to 
    contain 0.5*percent of all the weighted samples.
    If one of the two half-intervals does not have enough samples, it is considered
    un-defined. When this happens, the function consider it instead an upper/lower limit 
    on the sampled parameter and returns instead (None, upper-limit) or (lower-limit, None).
    
    Parameters
    ----------
    samples : ndarray or iterable
        the weighted samples over which to compute the estimator 
    idcentre : integer
        index in the samples iterable around which to compute the interval
    percent : float
        probability enclosed by the interval 
    weights :
        (Optional, default = None) weights of the samples, if not None
        this must have the same size of the samples argument
    
    Returns
    -------
    lower, upper : scalar floats
        [lower-, upper-) limits of the interval if the interval is defined.
        If it is not defined, the function will return either an upper limit or
        a lower limit, respectively.
    """
    
    if weights is None :
        weights = numpy.ones_like(samples)
    
    # sort the samples
    idsort = numpy.argsort(samples)
    
    # Normalize the weights to create a probability distribution
    weights = numpy.array(weights[idsort]) / numpy.sum(weights[idsort])

    # Calculate the cumulative distribution function (CDF)
    cdf = numpy.cumsum(weights)

    # Find the target value in the sorted array
    idx = numpy.nonzero(idsort == idcentre)

    # Find the lower bound based on the desired percentile
    lower_percentile = cdf[idx] - 0.5 * percent
    
    # account for upper limits
    if lower_percentile < 0.0 :
        upper_percentile = percent
        upper_idx = numpy.searchsorted(cdf, upper_percentile).item()
        return None, samples[idsort][upper_idx]
    
    # Find the upper bound based on the desired percentile
    upper_percentile = cdf[idx] + 0.5 * percent
    
    # account for lower limits
    if upper_percentile > 1.0 :
        lower_percentile = 1.0-percent
        lower_idx = numpy.searchsorted(cdf, lower_percentile).item()
        return samples[idsort][lower_idx], None

    # Find the lower and upper samples within the specified percentile range
    lower_idx = numpy.searchsorted(cdf, lower_percentile).item()
    upper_idx = numpy.searchsorted(cdf, upper_percentile).item()

    return samples[idsort][lower_idx], samples[idsort][upper_idx]

###################################################################################

def now_string () :
    """ Generates a string with the minute the function is called
    """
    from time import localtime as lt
    now = lt()
    return ( f'{now.tm_year:04}{now.tm_mon:02}{now.tm_mday:02}' +
             f'{now.tm_hour:02}{now.tm_min:02}' )

###################################################################################

def filter_strings ( inlist, fields ) :
    """Function for filtering a list of strings (uses ``fnmatch``)
    
    Parameters
    ----------
    inlist : sequence of str
        A list of strings to filter
    fields : str or sequence of str
        Either a single string or a sequence of strings. 
        Also accepts wildcards (e.g. ``which_params = 'sfh*'`` will show all the
        entries in ``inlist`` that contain the sub-string ``sfh``).
        
    Returns
    -------
    : str or sequence of str    
    
    Raises
    ------
    ValueError
    """
    
    import fnmatch
    
    if isinstance(fields, str) :
        fields = fnmatch.filter(inlist, fields)
    elif isinstance(fields, (list,tuple)) :
        fields = [item for sublist in [fnmatch.filter(inlist, s) for s in fields] for item in sublist]
    else :
        raise ValueError( 'fields argument should be either a string or a list of strings' )
        
    return fields
    
###################################################################################

def shorten_string ( string, maxlength = 10, splitchar = '.' ) :
    """Function for shortening a string, splitting it in components.
    
    Parameters
    ----------
    string : str
        Input string
    maxlength : int
        (default = 10) Maximum number of characters in the final string
    splitchar : str
        (default = '.') The character over which perform the splitting.
        
    Returns
    -------
    : str 
    The shortened string with length <= maxlength+len(splitchar)
    """
    
    components = string.split(splitchar)
    short_string = ''
    for component in numpy.flip(components) :
        if len(short_string) == 0 :
            short_string = component
        elif len(short_string) + len(component) + len(splitchar) < maxlength :
            short_string = splitchar.join([component, short_string])
        else :
            break
    return short_string
    
###################################################################################

def cat_to_dict ( infile, id_field = 'id', err_field = '_err', meta_fields = [], skip_fields = [] ) :
    """Converts a Topcat-like catalogue into a 2-levels dictionary.
    The 1st order dictionary contains a 2nd order dictionary for each entry in the catalogue.
    Each 2nd order dictionary contains data and meta-data about the named entry. 
    
    Parameters
    ----------
    infile : str
        Path to the csv file
    id_field : str
        (Optional, default='id') header name of the field containing the sources' ID
    err_field : str
        (Optional, default='_err') the sub-string identifying fields
        containing error measurements
    meta_fields : str or sequence of str
        (Optional, default empty list) which fields to consider meta-data
        Either a single string or a sequence of strings. 
        Also accepts wildcards (e.g. ``meta_fields = 'R*'`` will
        consider all the fields in header that start with ``R`` as meta-data).
    skip_fields : str or sequence of str
        (Optional, default empty list) which fields to skip
        Either a single string or a sequence of strings. 
        Also accepts wildcards (e.g. ``skip_fields = 'R*'`` will
        ignore all the fields in header that start with ``R``).
        
    
    Returns
    -------
    : dict
    """
    import re
    import warnings
    
    with open(infile, 'r') as f :
        lines = f.readlines()
    content = [ re.split( '\t+| +|,', line.rstrip() ) for line in lines ]
    
    gxydict = {}
    names = []
    for head, *gxys in zip(*content) :
        
        if head in filter_strings(content[0],skip_fields) :
            continue
        if head==id_field :
            names = gxys
            for name in names :
                gxydict[name] = {'bands':[], 'fluxes':[], 'errors':[]}
            continue
        for key, gxy in zip(names,gxys) :
            if numpy.any( [head==what 
                           for what in filter_strings(content[0], meta_fields)] ) :
                try :
                    gxydict[key][head] = float(gxy)
                except ValueError :
                    warnings.warn( f'{head} value {gxy} cannot be cast to float, skipping' )
                    continue
            elif err_field in head :
                try :
                    gxydict[key]['errors'].append(float(gxy))
                except ValueError :
                    continue
            else :
                try :
                    gxydict[key]['fluxes'].append(float(gxy))
                except ValueError :
                    continue
                gxydict[key]['bands'].append(head)

    for gxy in gxydict.values() :
        gxy['bands']  = numpy.asarray(gxy['bands'])
        gxy['fluxes'] = numpy.asarray(gxy['fluxes'])
        gxy['errors'] = numpy.asarray(gxy['errors'])
    
    return gxydict

###################################################################################

# def recurrent_return ( dd, keylist ) :
#     if len( keylist ) > 1 and isinstance( dd, MM ) :
#         return recurrent_return( dd[ keylist.pop( 0 ) ], keylist )
#     return dd[ keylist.pop( 0 ) ]

# def recurrent_update ( dd, **kwargs ) :
#     for k, v in kwargs.items() :
#         if isinstance(v,MM) :
#             rec_update( dd[k], **v )
#         elif k in dd.keys() :
#             dd[k] = v
#         else :
#             print( f'Key {k} not valid' )
#             continue
#     return;

###################################################################################

# class SmartAttribute ( object ) :
#     """ Descriptor object
    
#     Parameters
#     ----------
#     value : int, float, string, custom class, ... , optional
#         the value to set the described item at initialization
#         can be whathever python type, built-in or custom
#     valid : function 
#         a function that accepts a type( value ) object and
#         either returns it as is or raises a ValueError exception
#     """
    
#     def __init__ ( self, value=None, valid=None, string=str ) :
#         super( SmartAttribute, self ).__init__()
#         self.value = value
#         self.validate = valid
#         self.string = string
        
#     def __set__ ( self, obj, value ) :
#         """ Validates the input value before setting
#         """
#         self.value = self.validate( value )
        
#     def __get__ ( self, obj, objtype ) :
#         """ Returns the stored value
#         """
#         return self.value
    
#     def __repr__ ( self ) :
#         return f"{type(self).__name__}(value={self.value},valid={self.validate.__name__},string={self.string.__name__})"
    
#     def __str__ ( self ) :
#         return self.string( self.value )

###################################################################################

# class descAttrBase () :
#     """ Base class that allows described instance-attributes.
#     As described instance-attributes are parameters of the
#     derived classes, when one of them is changed it sets the
#     '_paramset' attribute to 'False'
#     (this is where the magic happens)
#     """

#     def __setattr__(self, attr, value):
#         try:
#             # Try invoking the descriptor protocol __set__ "manually"
#             got_attr = super().__getattribute__(attr)
#             got_attr.__set__(self, value)
#             # If class owns a '_paramset' attribute set it to False
#             # when changing one of the described attributes
#             try:
#                 _ = super().__getattribute__('_paramset')
#                 super().__setattr__('_paramset',False)
#             except AttributeError:
#                 pass
#         except AttributeError:
#             # Attribute is not a descriptor, just set it:
#             super().__setattr__(attr, value)

#     def __getattribute__(self, attr):
#         # If the attribute does not exist, super().__getattribute__()
#         # will raise an AttributeError
#         got_attr = super().__getattribute__(attr)
#         try:
#             # Try "manually" invoking the descriptor protocol __get__()
#             return got_attr.__get__(self, type(self))
#         except AttributeError:
#             # Attribute is not a descriptor, just return it:
#             return got_attr

###################################################################################

# class gxyComp ( descAttrBase, ABC ) :
#     """
#     """

#     def __init__ ( self ) :
#         self._paramset = False

#     def __bool__ ( self ) :
#         return self._paramset

#     def __and__ ( self, other ) :
#         return self._paramset.__and__( other )
#     def __or__ ( self, other ) :
#         return self._paramset.__or__( other )
#     # def __iand__ ( self, other ) :
#     #     return self._paramset.__and__( other )
#     # def __ior__ ( self, other ) :
#     #     return self._paramset.__or__( other )

#     @abstractmethod
#     def set_parameters ( self, *args, **kwargs ) :
#         pass

###################################################################################

# def set_nested(d, parent_obj):
#     """
#     Sets a collection of instance attributes stored in a nested dictionary
#     in the class instance parent_obj and in its sub-classes instances
#     recursevely.
    
#     Parameters
#     ----------
#     d : dictionary
#         Dictionary containing the instance attributes to be set.
#         In case of nested dictionaries the function will search the
#         class instance for attributes named as the key identifying 
#         the nested dictionary and eventually call itself recursively
#         to set its attributes.
#     parent_obj : class instance
#         the instance of a generic type where the new instance attributes 
#         will be set.
        
#     Examples
#     --------
#     >>> specs = { 'attr1' : 1, 'attr2' : 2 }
#     >>> class Foo () :
#     ...     pass
#     ... 
#     >>> foo = Foo()
#     >>> set_nested( specs, foo )
#     >>> foo.attr1
#     1
#     >>> foo.attr2
#     2
    
#     In case of nested classes
    
#     >>> nested_specs = { 'attr_bar' : 3, 'foo' : specs }
#     >>> class Bar () :
#     ...     def __init__ ( self ) :
#     ...             self.foo = Foo()
#     ... 
#     >>> bar = Bar()
#     >>> set_nested( nested_specs, bar )
#     >>> bar.attr_bar
#     3
#     >>> bar.foo.attr1, bar.foo.attr2
#     (1, 2)
#     """
#     for k, v in d.items():
#         if isinstance(v, collections.abc.MutableMapping):
#             set_nested( v, getattr( parent_obj, k ) )
#         else:
#             setattr( parent_obj, k, v )
#     return;

###################################################################################




import numpy
from scipy.special import erf

#############################################################################################

def simple_uplim ( data, error, model ) :
    data = numpy.asarray( data )
    error = numpy.asarray( error )
    model = numpy.asarray( model )
    
    if numpy.any( model > data ) :
        return -numpy.inf
    return 0.

#############################################################################################

# For upper limits (using eq. A10 of Sawicki2012):
def chi2_uplim ( data, error, model ) :
    
    # get argument of logarithm
    value = ( numpy.sqrt( 0.5 * numpy.pi ) *
              error *
              ( 1 + erf( numpy.sqrt( 0.5 ) *
                         ( data - model ) /
                         error ) ) )
    
    # check if there are values less than/equal to zero
    if numpy.any( ~( value > 0. ) ) :
        return numpy.inf

    # compute logarithm
    chi = numpy.log( value )

    return -2 * numpy.sum( chi )

#############################################################################################

def chi2 ( data, error, model ) :
    d, e, m = numpy.asarray(data), numpy.asarray(error), numpy.asarray(model)
    chi = ( d - m ) / e
    return numpy.sum( chi * chi )

#############################################################################################

f_uplims = { 'simple' : simple_uplim, 'chi2' : chi2, 'S12' : chi2_uplim }

#############################################################################################

def gaussian_loglikelihood ( data, error, model, uplims, method_uplims = 'simple' ) :
    """
    Uplim methods are:
    - 'simple' : 
    - 'chi2' :
    - 'S12' :
    """

    try :
        uplims = numpy.asarray( uplims )
    except KeyError as ke :
        raise RuntimeError( f"the method '{method_uplim}' is not valid, valid methods are {list(f_uplim.keys())}" )
    c2 = chi2( data[~uplims], error[~uplims], model[~uplims] )
    u2 = f_uplims[method_uplims]( data[uplims], error[uplims], model[uplims] )
    if numpy.isnan( c2 ) :
        return -numpy.inf
    if numpy.isnan( u2 ) :
        return -numpy.inf
        
    return -0.5 * ( c2 + u2 )

#############################################################################################

def transform_to_prior_unit_cube ( values, prior_limits ) :
    """ Maps the values of the parameters in the unit-cube prior, considered as uniform.
    
    Parameters
    ----------
    values : scalar or array-like
      value(s) of the parameters 
    prior_limits : array-like
      array with the lower/upper bounds of the uniform prior
      associated to each parameter listed in 'values'. 
      Shape should be Nx2 where N == len(values).

    Returns
    -------
    : scalar or array-like
    values mapped within the limits of the prior (all values in the [0.,1.) interval)
    """
    
    x = numpy.asarray( values ) 

    pmin, pmax = numpy.asarray( prior_limits ).T
    delta = pmax-pmin
    x *= delta
    x += pmin
        
    return x    

#############################################################################################



"""Implements the Nebular Continuum Emission (Bremsstrahlung) component.
"""

# External imports
import numpy

# Internal imports
from .NFF_core import CNFF

nff_tunables = ( 'Zgas', 'Zi' )

def nff_build_params ( **kwargs ) :
    
    out = {
        'Zgas' : 0.01,
        'Zi'   : 1.,
    }
    for k in set( out.keys() ).intersection( kwargs.keys() ) :
        out[ k ] = kwargs[ k ]

    return out

class NFF () :
    """ Class wrapping the C-core implementation of the Nebular Free-Free emission type.   
    
    Parameters
    ----------
    ll : float array
      the wavelength grid where the free-free emission is computed
    **kwargs : dictionary, optional
      All possible free-parameters of the model, refer to the
      Keyword Arguments section.
    
    Keyword Arguments
    -----------------
    Zgas : float
      (Optional, default = 0.01) Absolute metallicity of gas in the hosting galaxy.
    Zi : float
      (Optional, default = 1) Average atomic number of gas in the hosting galaxy.
    """
    
    def __init__ ( self, ll, **kwargs ) :

        self.l = numpy.ascontiguousarray( ll )
        self.core = CNFF( self.l )
        self.params = nff_build_params( **kwargs )
        self.set_parameters()

        
    def set_parameters ( self, **kwargs ) :
        r""" Function for setting the parameters of the model.

        Parameters
        ----------
        **kwargs : dictionary, optional
          All possible free-parameters of the model, refer to the
          Keyword Arguments section.
    
        Keyword Arguments
        -----------------
        Zgas : float
          Absolute metallicity of gas in the hosting galaxy.
        Zi : float
          Average atomic number of gas in the hosting galaxy.
        
        Returns
        -------
        : None
        """

        self.params.update( kwargs )
        self.core.set_params( numpy.asarray( [
            self.params[k]
            for k in nff_tunables
        ], dtype=float) )
        return;

    def gff ( self, il = None, Te = None ) :
        r"""Computes the Gaunt Factor as given in Draine, 2011 (Eq.10.9):

        .. math ::
            
            g_\text{NFF}=\ln\left\{\exp\left[5.96-\frac{\sqrt{3}}{\pi}\, 
            \ln\left( Z_i\, \frac{\nu_\lambda}{G\text{Hz}}\, \left(\frac{T_e}{10^4\, K}\right)^{-1.5}\right)\right]+
            \exp(1)\right\}
        
        where :math:`T_e` is the electron temperature (function argument), 
        :math:`Z_i` the average atomic number (class parameter) and 
        :math:`\nu_\lambda` is the frequency at given wavelength 
        (:math:`\lambda` as defined in the grid used to instantiate the class)
        
        Parameters
        ----------
        il : 1d array or scalar integer
          indices in the instance wavelength grid. If None, the gaunt factor
          is computed over all the wavelength grid.
        Te : float
          electron temperature of the medium, if None, the internal function :code:`NFF.Te()` is used.
        
        Returns
        -------
        : 1d array or scalar float
          Gaunt Factor at given wavelength
        """
        
        if il is None :
            il = numpy.arange( self.l.size, dtype = numpy.uint64 )
        il = numpy.asarray( il )
        scalar_input = False
        if il.ndim == 0 :
            il = il[None] # makes il 1D
            scalar_input = True
        if Te is None :
            Te = self.Te()

        ret = numpy.array( [ self.core.gff( _i, Te ) for _i in il ] )
        if scalar_input :
            return ret.item()
        return ret

    def Te ( self, Zgas = None ) :
        r"""Electron temperature as empirically derived in Vega et al., 2008:
        
        .. math ::
            
            T_e = \exp\left\{3.89-0.4802\, \log(Z_\text{gas}/0.02)-0.0205\,[\log(Z_\text{gas}/0.02)]^2\right\}
        
        where :math:`Z_\text{gas}` is the absolute gas metallicity.
        
        Parameters
        ----------
        Zgas : float
          (optional, default = same as parameters set) the average absolute metallicity of the gas.
          If None, it will use the internally stored parameter (can be set with :code:`NFF.set_parameters( Zgas = ... )`)
        
        Returns
        -------
        : float
          Electron temperature at given metallicity
        """
        
        if Zgas is None :
            Zgas = self.params[ 'Zgas' ]
        Zgas = numpy.asarray( Zgas )
        scalar_input = False
        if Zgas.ndim == 0 :
            Zgas = Zgas[None] # makes il 1D
            scalar_input = True
            
        ret = numpy.array( [ self.core.Te( _Z ) for _Z in Zgas ] )
        if scalar_input :
            return ret.item()
        return ret

    def emission ( self, Q_H, il = None ) :
        r"""Intrinsic Free-free emission as given in Draine, 2011 (Chapter 10):
        
        .. math::
            
            L_\text{NFF}(\lambda,\tau) \approx 1.8\times 10^{-27}\, \text{erg}\,s^{-1} \text{Hz}^{-1}\cdot
            \dfrac{\mathcal{Q}_\text{H}(\tau)}{s^{-1}}\, \left(\dfrac{T_e}{10^4\, K}\right)^{0.3}\, g_\text{NFF}\, 
            \exp\left(\dfrac{-h_P\, \nu_\lambda}{k_B\, T_e}\right)
        
        Parameters
        ----------
        Q_H : float
          intrinsic photo-ionisation rate, can be computed from the
          intrinsic stellar luminosity:
          
          .. math ::
              
              \mathcal{Q}_\text{H}(\tau) = \int_0^{912\mathring{A}}\text{d}\lambda\, 
              \frac{L_\text{CSP}^\text{i}(\lambda,\tau)}{h_P\, \nu_\lambda}
          
          where :math:`h_P` is the Planck constant and :math:`\nu_\lambda` is the frequency
          at given wavelength :math:`\lambda`.
          The intrinsic stellar luminosity, :math:`L_\text{CSP}^\text{i}(\lambda,\tau)`,
          can be computed using function :code:`galapy.CompositeStellarPopulation.CSP.emission()`.
        il : 1d array or scalar integer
          indices in the instance wavelength grid. If None, the emission
          is computed over all the wavelength grid.
        
        Returns
        -------
        : 1d array or scalar float
          intrinsic emission due to Free-Free transitions.
        """
        
        if il is None :
            il = numpy.arange( len(self.l), dtype = numpy.uint64 )
        il = numpy.asarray( il )
        scalar_input = False
        if il.ndim == 0 :
            il = il[None] # makes il 1D
            scalar_input = True
        
        ret = self.core.emission( il, Q_H )
        if scalar_input :
            return ret.item()
        return ret
        
    
